# type: ignore # this is riddled with typing issues -- fix this later.
import asyncio
import time
from contextlib import nullcontext
from pathlib import Path
from typing import Any, Awaitable, Callable, Dict, List, Optional, Set, Tuple, Union
from uuid import UUID

from ..._utils import Endpoint
from ...tracing import traced
from ..common import BaseService, FolderContext, UiPathApiConfig, UiPathExecutionContext
from ..errors import OperationFailedException, OperationNotCompleteException
from .documents import (
    ActionPriority,
    ClassificationResponse,
    ClassificationResult,
    ExtractionResponse,
    ExtractionResponseIXP,
    FileContent,
    ProjectType,
    StartExtractionResponse,
    StartOperationResponse,
    ValidateClassificationAction,
    ValidateExtractionAction,
)

POLLING_INTERVAL = 2  # seconds
POLLING_TIMEOUT = 300  # seconds


def _must_not_be_provided(**kwargs: Any) -> None:
    for name, value in kwargs.items():
        if value is not None:
            raise ValueError(f"`{name}` must not be provided")


def _must_be_provided(**kwargs: Any) -> None:
    for name, value in kwargs.items():
        if value is None:
            raise ValueError(f"`{name}` must be provided")


def _exactly_one_must_be_provided(**kwargs: Any) -> None:
    provided = [name for name, value in kwargs.items() if value is not None]
    if len(provided) != 1:
        raise ValueError(
            f"Exactly one of `{', '.join(kwargs.keys())}` must be provided"
        )


def _validate_classify_params(
    project_type: ProjectType,
    tag: Optional[str],
    project_name: Optional[str],
    file: Optional[FileContent],
    file_path: Optional[str],
) -> None:
    _exactly_one_must_be_provided(file=file, file_path=file_path)
    if project_type == ProjectType.PRETRAINED:
        _must_not_be_provided(
            project_name=project_name,
            tag=tag,
        )
    else:
        _must_be_provided(
            project_name=project_name,
            tag=tag,
        )


def _validate_extract_params_and_get_project_type(
    tag: Optional[str],
    project_name: Optional[str],
    file: Optional[FileContent],
    file_path: Optional[str],
    classification_result: Optional[ClassificationResult],
    project_type: Optional[ProjectType],
    document_type_name: Optional[str],
) -> ProjectType:
    if file or file_path:
        _exactly_one_must_be_provided(file=file, file_path=file_path)
        _must_be_provided(project_type=project_type)
        _must_not_be_provided(
            classification_result=classification_result,
        )

        if project_type == ProjectType.PRETRAINED:
            _must_not_be_provided(tag=tag, project_name=project_name)
            _must_be_provided(document_type_name=document_type_name)
        elif project_type == ProjectType.MODERN:
            _must_be_provided(
                project_name=project_name,
                tag=tag,
                document_type_name=document_type_name,
            )
        else:
            _must_be_provided(project_name=project_name, tag=tag)
            _must_not_be_provided(document_type_name=document_type_name)
    else:
        _must_be_provided(classification_result=classification_result)
        _must_not_be_provided(
            tag=tag,
            project_name=project_name,
            project_type=project_type,
            file=file,
            file_path=file_path,
            document_type_name=document_type_name,
        )
        project_type = classification_result.project_type

    return project_type


class DocumentsService(FolderContext, BaseService):
    """Service for managing UiPath DocumentUnderstanding Document Operations.

    This service provides methods to extract data from documents using UiPath's Document Understanding capabilities.

    !!! warning "Preview Feature"
        This function is currently experimental.
        Behavior and parameters are subject to change in future versions.
    """

    def __init__(
        self,
        config: UiPathApiConfig,
        execution_context: UiPathExecutionContext,
        polling_interval: float = POLLING_INTERVAL,
        polling_timeout: float = POLLING_TIMEOUT,
    ) -> None:
        super().__init__(config=config, execution_context=execution_context)
        self.polling_interval = polling_interval
        self.polling_timeout = polling_timeout

    def _get_common_headers(self) -> Dict[str, str]:
        return {
            "X-UiPath-Internal-ConsumptionSourceType": "CodedAgents",
        }

    def _get_project_id_by_name(
        self, project_name: str, project_type: ProjectType
    ) -> str:
        response = self.request(
            "GET",
            url=Endpoint("/du_/api/framework/projects"),
            params={"api-version": 1.1, "type": project_type.value},
            headers=self._get_common_headers(),
        )

        try:
            return next(
                project["id"]
                for project in response.json()["projects"]
                if project["name"] == project_name
            )
        except StopIteration:
            raise ValueError(f"Project '{project_name}' not found.") from None

    async def _get_project_id_by_name_async(
        self, project_name: str, project_type: ProjectType
    ) -> str:
        response = await self.request_async(
            "GET",
            url=Endpoint("/du_/api/framework/projects"),
            params={"api-version": 1.1, "type": project_type.value},
            headers=self._get_common_headers(),
        )

        try:
            return next(
                project["id"]
                for project in response.json()["projects"]
                if project["name"] == project_name
            )
        except StopIteration:
            raise ValueError(f"Project '{project_name}' not found.") from None

    def _get_project_tags(self, project_id: str) -> Set[str]:
        response = self.request(
            "GET",
            url=Endpoint(f"/du_/api/framework/projects/{project_id}/tags"),
            params={"api-version": 1.1},
            headers=self._get_common_headers(),
        )
        return {tag["name"] for tag in response.json().get("tags", [])}

    async def _get_project_tags_async(self, project_id: str) -> Set[str]:
        response = await self.request_async(
            "GET",
            url=Endpoint(f"/du_/api/framework/projects/{project_id}/tags"),
            params={"api-version": 1.1},
            headers=self._get_common_headers(),
        )
        return {tag["name"] for tag in response.json().get("tags", [])}

    def _get_document_id(
        self,
        project_id: Optional[str],
        file: Optional[FileContent],
        file_path: Optional[str],
        classification_result: Optional[ClassificationResult],
    ) -> str:
        if classification_result is not None:
            return classification_result.document_id

        document_id = self._start_digitization(
            project_id=project_id,
            file=file,
            file_path=file_path,
        )
        self._wait_for_digitization(
            project_id=project_id,
            document_id=document_id,
        )

        return document_id

    async def _get_document_id_async(
        self,
        project_id: Optional[str],
        file: Optional[FileContent],
        file_path: Optional[str],
        classification_result: Optional[ClassificationResult],
    ) -> str:
        if classification_result is not None:
            return classification_result.document_id

        document_id = await self._start_digitization_async(
            project_id=project_id,
            file=file,
            file_path=file_path,
        )
        await self._wait_for_digitization_async(
            project_id=project_id,
            document_id=document_id,
        )

        return document_id

    def _get_project_id_and_tag(
        self,
        tag: Optional[str],
        project_name: Optional[str],
        project_type: Optional[ProjectType],
        classification_result: Optional[ClassificationResult],
    ) -> Tuple[str, Optional[str]]:
        if project_name is not None:
            project_id = self._get_project_id_by_name(
                project_name,
                project_type,
            )
        elif project_type == ProjectType.PRETRAINED:
            return str(UUID(int=0)), None
        else:
            project_id = classification_result.project_id
            tag = classification_result.tag

        tags = self._get_project_tags(project_id)
        if tag not in tags:
            raise ValueError(
                f"Tag '{tag}' not found in project '{project_name}'. Available tags: {tags}"
            )

        return project_id, tag

    async def _get_project_id_and_tag_async(
        self,
        tag: Optional[str],
        project_name: Optional[str],
        project_type: Optional[ProjectType],
        classification_result: Optional[ClassificationResult],
    ) -> Tuple[str, Optional[str]]:
        if project_name is not None:
            project_id = await self._get_project_id_by_name_async(
                project_name,
                project_type,
            )
        elif project_type == ProjectType.PRETRAINED:
            return str(UUID(int=0)), None
        else:
            project_id = classification_result.project_id
            tag = classification_result.tag

        tags = await self._get_project_tags_async(project_id)
        if tag not in tags:
            raise ValueError(
                f"Tag '{tag}' not found in project '{project_name}'. Available tags: {tags}"
            )

        return project_id, tag

    def _start_digitization(
        self,
        project_id: str,
        file: Optional[FileContent] = None,
        file_path: Optional[str] = None,
    ) -> str:
        with open(Path(file_path), "rb") if file_path else nullcontext(file) as handle:
            return self.request(
                "POST",
                url=Endpoint(
                    f"/du_/api/framework/projects/{project_id}/digitization/start"
                ),
                params={"api-version": 1.1},
                headers=self._get_common_headers(),
                files={"File": handle},
            ).json()["documentId"]

    async def _start_digitization_async(
        self,
        project_id: str,
        file: Optional[FileContent] = None,
        file_path: Optional[str] = None,
    ) -> str:
        with open(Path(file_path), "rb") if file_path else nullcontext(file) as handle:
            return (
                await self.request_async(
                    "POST",
                    url=Endpoint(
                        f"/du_/api/framework/projects/{project_id}/digitization/start"
                    ),
                    params={"api-version": 1.1},
                    headers=self._get_common_headers(),
                    files={"File": handle},
                )
            ).json()["documentId"]

    def _wait_for_digitization(self, project_id: str, document_id: str) -> None:
        def result_getter() -> Tuple[str, Optional[str], Optional[str]]:
            result = self.request(
                method="GET",
                url=Endpoint(
                    f"/du_/api/framework/projects/{project_id}/digitization/result/{document_id}"
                ),
                params={"api-version": 1.1},
                headers=self._get_common_headers(),
            ).json()
            return (
                result["status"],
                result.get("error", None),
                result.get("result", None),
            )

        self._wait_for_operation(
            result_getter=result_getter,
            wait_statuses=["NotStarted", "Running"],
            success_status="Succeeded",
        )

    async def _wait_for_digitization_async(
        self, project_id: str, document_id: str
    ) -> None:
        async def result_getter() -> Tuple[str, Optional[str], Optional[str]]:
            result = (
                await self.request_async(
                    method="GET",
                    url=Endpoint(
                        f"/du_/api/framework/projects/{project_id}/digitization/result/{document_id}"
                    ),
                    params={"api-version": 1.1},
                    headers=self._get_common_headers(),
                )
            ).json()
            return (
                result["status"],
                result.get("error", None),
                result.get("result", None),
            )

        await self._wait_for_operation_async(
            result_getter=result_getter,
            wait_statuses=["NotStarted", "Running"],
            success_status="Succeeded",
        )

    def _get_document_type_id(
        self,
        project_id: str,
        document_type_name: Optional[str],
        project_type: ProjectType,
        classification_result: Optional[ClassificationResult],
    ) -> str:
        if project_type == ProjectType.IXP:
            return str(UUID(int=0))

        if classification_result is not None:
            return classification_result.document_type_id

        response = self.request(
            "GET",
            url=Endpoint(f"/du_/api/framework/projects/{project_id}/document-types"),
            params={"api-version": 1.1},
            headers=self._get_common_headers(),
        )

        try:
            return next(
                extractor["id"]
                for extractor in response.json().get("documentTypes", [])
                if extractor["name"].lower() == document_type_name.lower()
            )
        except StopIteration:
            raise ValueError(
                f"Document type '{document_type_name}' not found."
            ) from None

    async def _get_document_type_id_async(
        self,
        project_id: str,
        document_type_name: Optional[str],
        project_type: ProjectType,
        classification_result: Optional[ClassificationResult],
    ) -> str:
        if project_type == ProjectType.IXP:
            return str(UUID(int=0))

        if classification_result is not None:
            return classification_result.document_type_id

        response = await self.request_async(
            "GET",
            url=Endpoint(f"/du_/api/framework/projects/{project_id}/document-types"),
            params={"api-version": 1.1},
            headers=self._get_common_headers(),
        )

        try:
            return next(
                extractor["id"]
                for extractor in response.json().get("documentTypes", [])
                if extractor["name"].lower() == document_type_name.lower()
            )
        except StopIteration:
            raise ValueError(
                f"Document type '{document_type_name}' not found."
            ) from None

    def _start_extraction(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        document_type_id: str,
        document_id: str,
    ) -> StartExtractionResponse:
        if project_type == ProjectType.PRETRAINED:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/extractors/{document_type_id}/extraction/start"
            )
        else:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/{tag}/document-types/{document_type_id}/extraction/start"
            )

        operation_id = self.request(
            "POST",
            url=url,
            params={"api-version": 1.1},
            headers=self._get_common_headers(),
            json={"documentId": document_id},
        ).json()["operationId"]

        return StartExtractionResponse(
            operation_id=operation_id,
            document_id=document_id,
            project_id=project_id,
            tag=tag,
        )

    async def _start_extraction_async(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        document_type_id: str,
        document_id: str,
    ) -> StartExtractionResponse:
        if project_type == ProjectType.PRETRAINED:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/extractors/{document_type_id}/extraction/start"
            )
        else:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/{tag}/document-types/{document_type_id}/extraction/start"
            )

        operation_id = (
            await self.request_async(
                "POST",
                url=url,
                params={"api-version": 1.1},
                headers=self._get_common_headers(),
                json={"documentId": document_id},
            )
        ).json()["operationId"]

        return StartExtractionResponse(
            operation_id=operation_id,
            document_id=document_id,
            project_id=project_id,
            tag=tag,
        )

    def _wait_for_operation(
        self,
        result_getter: Callable[[], Tuple[Any, Optional[Any], Optional[Any]]],
        wait_statuses: List[str],
        success_status: str,
    ) -> Any:
        start_time = time.monotonic()
        status = wait_statuses[0]
        result = None

        while (
            status in wait_statuses
            and (time.monotonic() - start_time) < self.polling_timeout
        ):
            status, error, result = result_getter()
            time.sleep(self.polling_interval)

        if status != success_status:
            if time.monotonic() - start_time >= self.polling_timeout:
                raise TimeoutError("Operation timed out.")
            raise RuntimeError(
                f"Operation failed with status: {status}, error: {error}"
            )

        return result

    async def _wait_for_operation_async(
        self,
        result_getter: Callable[
            [], Awaitable[Tuple[Any, Optional[Any], Optional[Any]]]
        ],
        wait_statuses: List[str],
        success_status: str,
    ) -> Any:
        start_time = time.monotonic()
        status = wait_statuses[0]
        result = None

        while (
            status in wait_statuses
            and (time.monotonic() - start_time) < self.polling_timeout
        ):
            status, error, result = await result_getter()
            await asyncio.sleep(self.polling_interval)

        if status != success_status:
            if time.monotonic() - start_time >= self.polling_timeout:
                raise TimeoutError("Operation timed out.")
            raise RuntimeError(
                f"Operation failed with status: {status}, error: {error}"
            )

        return result

    def _wait_for_extraction(
        self,
        project_id: str,
        tag: Optional[str],
        document_type_id: str,
        operation_id: str,
        project_type: ProjectType,
    ) -> Union[ExtractionResponse, ExtractionResponseIXP]:
        def result_getter() -> Tuple[str, str, Any]:
            if project_type == ProjectType.PRETRAINED:
                url = Endpoint(
                    f"/du_/api/framework/projects/{project_id}/extractors/{document_type_id}/extraction/result/{operation_id}"
                )
            else:
                url = Endpoint(
                    f"/du_/api/framework/projects/{project_id}/{tag}/document-types/{document_type_id}/extraction/result/{operation_id}"
                )

            result = self.request(
                method="GET",
                url=url,
                params={"api-version": 1.1},
                headers=self._get_common_headers(),
            ).json()
            return (
                result["status"],
                result.get("error", None),
                result.get("result", None),
            )

        extraction_response = self._wait_for_operation(
            result_getter=result_getter,
            wait_statuses=["NotStarted", "Running"],
            success_status="Succeeded",
        )

        extraction_response["projectId"] = project_id
        extraction_response["tag"] = tag
        extraction_response["documentTypeId"] = document_type_id
        extraction_response["projectType"] = project_type

        if project_type == ProjectType.IXP:
            return ExtractionResponseIXP.model_validate(extraction_response)

        return ExtractionResponse.model_validate(extraction_response)

    async def _wait_for_extraction_async(
        self,
        project_id: str,
        tag: Optional[str],
        document_type_id: str,
        operation_id: str,
        project_type: ProjectType,
    ) -> Union[ExtractionResponse, ExtractionResponseIXP]:
        async def result_getter() -> Tuple[str, str, Any]:
            if project_type == ProjectType.PRETRAINED:
                url = Endpoint(
                    f"/du_/api/framework/projects/{project_id}/extractors/{document_type_id}/extraction/result/{operation_id}"
                )
            else:
                url = Endpoint(
                    f"/du_/api/framework/projects/{project_id}/{tag}/document-types/{document_type_id}/extraction/result/{operation_id}"
                )

            result = (
                await self.request_async(
                    method="GET",
                    url=url,
                    params={"api-version": 1.1},
                    headers=self._get_common_headers(),
                )
            ).json()
            return (
                result["status"],
                result.get("error", None),
                result.get("result", None),
            )

        extraction_response = await self._wait_for_operation_async(
            result_getter=result_getter,
            wait_statuses=["NotStarted", "Running"],
            success_status="Succeeded",
        )

        extraction_response["projectId"] = project_id
        extraction_response["tag"] = tag
        extraction_response["documentTypeId"] = document_type_id
        extraction_response["projectType"] = project_type

        if project_type == ProjectType.IXP:
            return ExtractionResponseIXP.model_validate(extraction_response)

        return ExtractionResponse.model_validate(extraction_response)

    def _start_classification(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        document_id: str,
    ) -> str:
        if project_type == ProjectType.PRETRAINED:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/classifiers/ml-classification/classification/start"
            )
        else:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/{tag}/classification/start"
            )

        return self.request(
            "POST",
            url=url,
            params={"api-version": 1.1},
            headers=self._get_common_headers(),
            json={"documentId": document_id},
        ).json()["operationId"]

    async def _start_classification_async(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        document_id: str,
    ) -> str:
        if project_type == ProjectType.PRETRAINED:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/classifiers/ml-classification/classification/start"
            )
        else:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/{tag}/classification/start"
            )

        return (
            await self.request_async(
                "POST",
                url=url,
                params={"api-version": 1.1},
                headers=self._get_common_headers(),
                json={"documentId": document_id},
            )
        ).json()["operationId"]

    def _wait_for_classification(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        operation_id: str,
    ) -> List[ClassificationResult]:
        def result_getter() -> Tuple[str, Optional[str], Optional[str]]:
            if project_type == ProjectType.PRETRAINED:
                url = Endpoint(
                    f"/du_/api/framework/projects/{project_id}/classifiers/ml-classification/classification/result/{operation_id}"
                )
            else:
                url = Endpoint(
                    f"/du_/api/framework/projects/{project_id}/{tag}/classification/result/{operation_id}"
                )

            result = self.request(
                method="GET",
                url=url,
                params={"api-version": 1.1},
                headers=self._get_common_headers(),
            ).json()
            return (
                result["status"],
                result.get("error", None),
                result.get("result", None),
            )

        classification_response = self._wait_for_operation(
            result_getter=result_getter,
            wait_statuses=["NotStarted", "Running"],
            success_status="Succeeded",
        )
        for classification_result in classification_response["classificationResults"]:
            classification_result["ProjectId"] = project_id
            classification_result["ProjectType"] = project_type
            classification_result["Tag"] = tag

        return ClassificationResponse.model_validate(
            classification_response
        ).classification_results

    async def _wait_for_classification_async(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        operation_id: str,
    ) -> List[ClassificationResult]:
        async def result_getter() -> Tuple[str, Optional[str], Optional[str]]:
            if project_type == ProjectType.PRETRAINED:
                url = Endpoint(
                    f"/du_/api/framework/projects/{project_id}/classifiers/ml-classification/classification/result/{operation_id}"
                )
            else:
                url = Endpoint(
                    f"/du_/api/framework/projects/{project_id}/{tag}/classification/result/{operation_id}"
                )

            result = (
                await self.request_async(
                    method="GET",
                    url=url,
                    params={"api-version": 1.1},
                    headers=self._get_common_headers(),
                )
            ).json()
            return (
                result["status"],
                result.get("error", None),
                result.get("result", None),
            )

        classification_response = await self._wait_for_operation_async(
            result_getter=result_getter,
            wait_statuses=["NotStarted", "Running"],
            success_status="Succeeded",
        )
        for classification_result in classification_response["classificationResults"]:
            classification_result["ProjectId"] = project_id
            classification_result["ProjectType"] = project_type
            classification_result["Tag"] = tag

        return ClassificationResponse.model_validate(
            classification_response
        ).classification_results

    @traced(name="documents_classify", run_type="uipath")
    def classify(
        self,
        project_type: ProjectType,
        tag: Optional[str] = None,
        project_name: Optional[str] = None,
        file: Optional[FileContent] = None,
        file_path: Optional[str] = None,
    ) -> List[ClassificationResult]:
        """Classify a document using a DU Modern project.

        Args:
            project_type (ProjectType): Type of the project.
            project_name (str, optional): Name of the [DU Modern](https://docs.uipath.com/document-understanding/automation-cloud/latest/user-guide/about-document-understanding) project. Must be provided if `project_type` is not `ProjectType.PRETRAINED`.
            tag (str, optional): Tag of the published project version. Must be provided if `project_type` is not `ProjectType.PRETRAINED`.
            file (FileContent, optional): The document file to be classified.
            file_path (str, optional): Path to the document file to be classified.

        Note:
            Either `file` or `file_path` must be provided, but not both.

        Returns:
            List[ClassificationResult]: A list of classification results.

        Examples:
            ```python
            Modern DU project:
            with open("path/to/document.pdf", "rb") as file:
                classification_results = service.classify(
                    project_name="MyModernProjectName",
                    tag="Production",
                    file=file,
                )

            Pretrained project:
            with open("path/to/document.pdf", "rb") as file:
                classification_results = service.classify(
                    project_type=ProjectType.PRETRAINED,
                    file=file,
                )
            ```
        """
        _validate_classify_params(
            project_type=project_type,
            tag=tag,
            project_name=project_name,
            file=file,
            file_path=file_path,
        )

        project_id, tag = self._get_project_id_and_tag(
            tag=tag,
            project_name=project_name,
            project_type=project_type,
            classification_result=None,
        )

        document_id = self._get_document_id(
            project_id=project_id,
            file=file,
            file_path=file_path,
            classification_result=None,
        )

        operation_id = self._start_classification(
            project_id=project_id,
            project_type=project_type,
            tag=tag,
            document_id=document_id,
        )
        return self._wait_for_classification(
            project_id=project_id,
            project_type=project_type,
            tag=tag,
            operation_id=operation_id,
        )

    @traced(name="documents_classify_async", run_type="uipath")
    async def classify_async(
        self,
        project_type: ProjectType,
        tag: Optional[str] = None,
        project_name: Optional[str] = None,
        file: Optional[FileContent] = None,
        file_path: Optional[str] = None,
    ) -> List[ClassificationResult]:
        """Asynchronously version of the [`classify`][uipath.platform.documents._documents_service.DocumentsService.classify] method."""
        _validate_classify_params(
            project_type=project_type,
            tag=tag,
            project_name=project_name,
            file=file,
            file_path=file_path,
        )

        project_id, tag = await self._get_project_id_and_tag_async(
            tag=tag,
            project_name=project_name,
            project_type=project_type,
            classification_result=None,
        )

        document_id = await self._get_document_id_async(
            project_id=project_id,
            file=file,
            file_path=file_path,
            classification_result=None,
        )

        operation_id = await self._start_classification_async(
            project_id=project_id,
            project_type=project_type,
            tag=tag,
            document_id=document_id,
        )

        return await self._wait_for_classification_async(
            project_id=project_id,
            project_type=project_type,
            tag=tag,
            operation_id=operation_id,
        )

    @traced(name="documents_start_ixp_extraction", run_type="uipath")
    def start_ixp_extraction(
        self,
        project_name: str,
        tag: str,
        file: Optional[FileContent] = None,
        file_path: Optional[str] = None,
    ) -> StartExtractionResponse:
        """Start an IXP extraction process without waiting for results (non-blocking).

        This method uploads the file as an attachment and starts the extraction process,
        returning immediately without waiting for the extraction to complete.
        Use this for async workflows where you want to receive results via callback/webhook.

        Args:
            project_name (str): Name of the IXP project.
            tag (str): Tag of the published project version (e.g., "staging").
            file (FileContent, optional): The document file to be processed.
            file_path (str, optional): Path to the document file to be processed.

        Note:
            Either `file` or `file_path` must be provided, but not both.

        Returns:
            ExtractionStartResponse: Contains the operation_id, document_id, project_id, and tag

        Examples:
            ```python
            start_response = uipath.documents.start_ixp_extraction(
                project_name="MyIXPProjectName",
                tag="staging",
                file_path="path/to/document.pdf",
            )
            # start_response.operation_id can be used to poll for results later
            ```
        """
        _exactly_one_must_be_provided(file=file, file_path=file_path)

        project_id = self._get_project_id_by_name(project_name, ProjectType.IXP)

        document_id = self._start_digitization(
            project_id=project_id,
            file=file,
            file_path=file_path,
        )

        return self._start_extraction(
            project_id=project_id,
            project_type=ProjectType.IXP,
            tag=tag,
            document_type_id=str(UUID(int=0)),
            document_id=document_id,
        )

    @traced(name="documents_start_ixp_extraction_async", run_type="uipath")
    async def start_ixp_extraction_async(
        self,
        project_name: str,
        tag: str,
        file: Optional[FileContent] = None,
        file_path: Optional[str] = None,
    ) -> StartExtractionResponse:
        """Asynchronous version of the [`start_ixp_extraction`][uipath.platform.documents._documents_service.DocumentsService.start_ixp_extraction] method."""
        _exactly_one_must_be_provided(file=file, file_path=file_path)

        project_id = await self._get_project_id_by_name_async(
            project_name, ProjectType.IXP
        )

        document_id = await self._start_digitization_async(
            project_id=project_id,
            file=file,
            file_path=file_path,
        )

        return await self._start_extraction_async(
            project_id=project_id,
            project_type=ProjectType.IXP,
            tag=tag,
            document_type_id=str(UUID(int=0)),
            document_id=document_id,
        )

    def _retrieve_operation_result(
        self,
        url: Endpoint,
        operation_id: str,
        operation_name: str,
    ) -> Dict:
        response = self.request(
            method="GET",
            url=url,
            params={"api-version": "1.1"},
            headers=self._get_common_headers(),
        ).json()

        status = response.get("status")
        if status in ["NotStarted", "Running"]:
            raise OperationNotCompleteException(
                operation_id=operation_id,
                status=response.get("status"),
                operation_name=operation_name,
            )

        if status != "Succeeded":
            raise OperationFailedException(
                operation_id=operation_id,
                status=status,
                error=response.get("error"),
                operation_name=operation_name,
            )

        return response.get("result")

    async def _retrieve_operation_result_async(
        self,
        url: Endpoint,
        operation_id: str,
        operation_name: str,
    ) -> Dict:
        response = (
            await self.request_async(
                method="GET",
                url=url,
                params={"api-version": "1.1"},
                headers=self._get_common_headers(),
            )
        ).json()

        status = response.get("status")
        if status in ["NotStarted", "Running"]:
            raise OperationNotCompleteException(
                operation_id=operation_id,
                status=response.get("status"),
                operation_name=operation_name,
            )

        if status != "Succeeded":
            raise OperationFailedException(
                operation_id=operation_id,
                status=status,
                error=response.get("error"),
                operation_name=operation_name,
            )

        return response.get("result")

    @traced(name="documents_retrieve_ixp_extraction_result", run_type="uipath")
    def retrieve_ixp_extraction_result(
        self,
        project_id: str,
        tag: str,
        operation_id: str,
    ) -> ExtractionResponseIXP:
        """Retrieve the result of an IXP extraction operation (single-shot, non-blocking).

        This method retrieves the result of an IXP extraction that was previously started
        with `start_ixp_extraction`. It does not poll - it makes a single request and
        returns the result if available, or raises an exception if not complete.

        Args:
            project_id (str): The ID of the IXP project.
            tag (str): The tag of the published project version.
            operation_id (str): The operation ID returned from `start_ixp_extraction`.

        Returns:
            ExtractionResponseIXP: The extraction response containing the extracted data.

        Raises:
            OperationNotCompleteException: If the extraction is not yet complete.
            OperationFailedException: If the extraction operation failed.

        Examples:
            ```python
            # After receiving a callback/webhook that extraction is complete:
            result = service.retrieve_ixp_extraction_result(
                project_id=start_response.project_id,
                tag=start_response.tag,
                operation_id=start_response.operation_id,
            )
            ```
        """
        document_type_id = str(UUID(int=0))

        url = Endpoint(
            f"/du_/api/framework/projects/{project_id}/{tag}/document-types/{document_type_id}/extraction/result/{operation_id}"
        )

        extraction_response = self._retrieve_operation_result(
            url=url,
            operation_id=operation_id,
            operation_name="IXP extraction",
        )

        extraction_response["projectId"] = project_id
        extraction_response["tag"] = tag
        extraction_response["documentTypeId"] = document_type_id
        extraction_response["projectType"] = ProjectType.IXP

        return ExtractionResponseIXP.model_validate(extraction_response)

    @traced(name="documents_retrieve_ixp_extraction_result_async", run_type="uipath")
    async def retrieve_ixp_extraction_result_async(
        self,
        project_id: str,
        tag: str,
        operation_id: str,
    ) -> ExtractionResponseIXP:
        """Asynchronous version of the [`retrieve_ixp_extraction_result`][uipath.platform.documents._documents_service.DocumentsService.retrieve_ixp_extraction_result] method."""
        document_type_id = str(UUID(int=0))

        url = Endpoint(
            f"/du_/api/framework/projects/{project_id}/{tag}/document-types/{document_type_id}/extraction/result/{operation_id}"
        )

        extraction_response = await self._retrieve_operation_result_async(
            url=url,
            operation_id=operation_id,
            operation_name="IXP extraction",
        )

        extraction_response["projectId"] = project_id
        extraction_response["tag"] = tag
        extraction_response["documentTypeId"] = document_type_id
        extraction_response["projectType"] = ProjectType.IXP

        return ExtractionResponseIXP.model_validate(extraction_response)

    @traced(name="documents_extract", run_type="uipath")
    def extract(
        self,
        tag: Optional[str] = None,
        project_name: Optional[str] = None,
        file: Optional[FileContent] = None,
        file_path: Optional[str] = None,
        classification_result: Optional[ClassificationResult] = None,
        project_type: Optional[ProjectType] = None,
        document_type_name: Optional[str] = None,
    ) -> Union[ExtractionResponse, ExtractionResponseIXP]:
        """Extract predicted data from a document using an DU Modern/IXP project.

        Args:
            project_name (str, optional): Name of the [IXP](https://docs.uipath.com/ixp/automation-cloud/latest/overview/managing-projects#creating-a-new-project)/[DU Modern](https://docs.uipath.com/document-understanding/automation-cloud/latest/user-guide/about-document-understanding) project. Must be provided if `classification_result` is not provided.
            tag (str): Tag of the published project version. Must be provided if `classification_result` is not provided and `project_type` is not `ProjectType.PRETRAINED`.
            file (FileContent, optional): The document file to be processed. Must be provided if `classification_result` is not provided.
            file_path (str, optional): Path to the document file to be processed. Must be provided if `classification_result` is not provided.
            project_type (ProjectType, optional): Type of the project. Must be provided if `project_name` is provided.
            document_type_name (str, optional): Document type name associated with the extractor to be used for extraction. Required if `project_type` is `ProjectType.MODERN` and `project_name` is provided.
            classification_result (ClassificationResult, optional): The classification result obtained from a previous classification step. If provided, `project_name`, `project_type`, `file`, `file_path`, and `document_type_name` must not be provided.

        Note:
            Either `file` or `file_path` must be provided, but not both.

        Returns:
            Union[ExtractionResponse, ExtractionResponseIXP]: The extraction response containing the extracted data.

        Examples:
            IXP projects:
            ```python
            with open("path/to/document.pdf", "rb") as file:
                extraction_response = service.extract(
                    project_name="MyIXPProjectName",
                    tag="live",
                    file=file,
                )
            ```

            DU Modern projects (providing document type name):
            ```python
            with open("path/to/document.pdf", "rb") as file:
                extraction_response = service.extract(
                    project_name="MyModernProjectName",
                    tag="Production",
                    file=file,
                    project_type=ProjectType.MODERN,
                    document_type_name="Receipts",
                )
            ```

            DU Modern projects (using existing classification result):
            ```python
            with open("path/to/document.pdf", "rb") as file:
                classification_results = uipath.documents.classify(
                    tag="Production",
                    project_name="MyModernProjectName",
                    file=file,
                )

            extraction_result = uipath.documents.extract(
                classification_result=max(classification_results, key=lambda result: result.confidence),
            )
            ```
        """
        project_type = _validate_extract_params_and_get_project_type(
            tag=tag,
            project_name=project_name,
            file=file,
            file_path=file_path,
            classification_result=classification_result,
            project_type=project_type,
            document_type_name=document_type_name,
        )

        project_id, tag = self._get_project_id_and_tag(
            tag=tag,
            project_name=project_name,
            project_type=project_type,
            classification_result=classification_result,
        )

        document_id = self._get_document_id(
            project_id=project_id,
            file=file,
            file_path=file_path,
            classification_result=classification_result,
        )

        document_type_id = self._get_document_type_id(
            project_id=project_id,
            document_type_name=document_type_name,
            project_type=project_type,
            classification_result=classification_result,
        )

        operation_id = self._start_extraction(
            project_id=project_id,
            project_type=project_type,
            tag=tag,
            document_type_id=document_type_id,
            document_id=document_id,
        ).operation_id

        return self._wait_for_extraction(
            project_id=project_id,
            tag=tag,
            document_type_id=document_type_id,
            operation_id=operation_id,
            project_type=project_type,
        )

    @traced(name="documents_extract_async", run_type="uipath")
    async def extract_async(
        self,
        tag: Optional[str] = None,
        project_name: Optional[str] = None,
        file: Optional[FileContent] = None,
        file_path: Optional[str] = None,
        classification_result: Optional[ClassificationResult] = None,
        project_type: Optional[ProjectType] = None,
        document_type_name: Optional[str] = None,
    ) -> Union[ExtractionResponse, ExtractionResponseIXP]:
        """Asynchronously version of the [`extract`][uipath.platform.documents._documents_service.DocumentsService.extract] method."""
        project_type = _validate_extract_params_and_get_project_type(
            tag=tag,
            project_name=project_name,
            file=file,
            file_path=file_path,
            classification_result=classification_result,
            project_type=project_type,
            document_type_name=document_type_name,
        )

        project_id, tag = await self._get_project_id_and_tag_async(
            tag=tag,
            project_name=project_name,
            project_type=project_type,
            classification_result=classification_result,
        )

        document_id = await self._get_document_id_async(
            project_id=project_id,
            file=file,
            file_path=file_path,
            classification_result=classification_result,
        )

        document_type_id = await self._get_document_type_id_async(
            project_id=project_id,
            document_type_name=document_type_name,
            project_type=project_type,
            classification_result=classification_result,
        )

        operation_id = (
            await self._start_extraction_async(
                project_id=project_id,
                project_type=project_type,
                tag=tag,
                document_type_id=document_type_id,
                document_id=document_id,
            )
        ).operation_id

        return await self._wait_for_extraction_async(
            project_id=project_id,
            tag=tag,
            document_type_id=document_type_id,
            operation_id=operation_id,
            project_type=project_type,
        )

    def _start_classification_validation(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        classification_results: List[ClassificationResult],
        action_title: str,
        action_priority: Optional[ActionPriority] = None,
        action_catalog: Optional[str] = None,
        action_folder: Optional[str] = None,
        storage_bucket_name: Optional[str] = None,
        storage_bucket_directory_path: Optional[str] = None,
    ) -> str:
        if project_type == ProjectType.PRETRAINED:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/classifiers/ml-classification/validation/start"
            )
        else:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/{tag}/classifiers/validation/start"
            )

        return self.request(
            "POST",
            url=url,
            params={"api-version": 1.1},
            headers=self._get_common_headers(),
            json={
                "classificationResults": [
                    cr.model_dump() for cr in classification_results
                ],
                "documentId": classification_results[0].document_id,
                "actionTitle": action_title,
                "actionPriority": action_priority,
                "actionCatalog": action_catalog,
                "actionFolder": action_folder,
                "storageBucketName": storage_bucket_name,
                "storageBucketDirectoryPath": storage_bucket_directory_path,
            },
        ).json()["operationId"]

    async def _start_classification_validation_async(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        classification_results: List[ClassificationResult],
        action_title: str,
        action_priority: Optional[ActionPriority] = None,
        action_catalog: Optional[str] = None,
        action_folder: Optional[str] = None,
        storage_bucket_name: Optional[str] = None,
        storage_bucket_directory_path: Optional[str] = None,
    ) -> str:
        if project_type == ProjectType.PRETRAINED:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/classifiers/ml-classification/validation/start"
            )
        else:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/{tag}/classifiers/validation/start"
            )

        return (
            await self.request_async(
                "POST",
                url=url,
                params={"api-version": 1.1},
                headers=self._get_common_headers(),
                json={
                    "classificationResults": [
                        cr.model_dump() for cr in classification_results
                    ],
                    "documentId": classification_results[0].document_id,
                    "actionTitle": action_title,
                    "actionPriority": action_priority,
                    "actionCatalog": action_catalog,
                    "actionFolder": action_folder,
                    "storageBucketName": storage_bucket_name,
                    "storageBucketDirectoryPath": storage_bucket_directory_path,
                },
            )
        ).json()["operationId"]

    def _start_extraction_validation(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        document_type_id: str,
        action_title: str,
        action_priority: Optional[ActionPriority],
        action_catalog: Optional[str],
        action_folder: Optional[str],
        storage_bucket_name: Optional[str],
        storage_bucket_directory_path: Optional[str],
        extraction_response: ExtractionResponse,
    ) -> StartOperationResponse:
        if project_type == ProjectType.PRETRAINED:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/extractors/{document_type_id}/validation/start"
            )
        else:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/{tag}/document-types/{document_type_id}/validation/start"
            )

        operation_id = self.request(
            "POST",
            url=url,
            params={"api-version": 1.1},
            headers=self._get_common_headers(),
            json={
                "extractionResult": extraction_response.extraction_result.model_dump(),
                "documentId": extraction_response.extraction_result.document_id,
                "actionTitle": action_title,
                "actionPriority": action_priority,
                "actionCatalog": action_catalog,
                "actionFolder": action_folder,
                "storageBucketName": storage_bucket_name,
                "allowChangeOfDocumentType": True,
                "storageBucketDirectoryPath": storage_bucket_directory_path,
            },
        ).json()["operationId"]

        return StartOperationResponse(
            operation_id=operation_id,
            document_id=extraction_response.extraction_result.document_id,
            project_id=project_id,
            tag=tag,
        )

    async def _start_extraction_validation_async(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        document_type_id: str,
        action_title: str,
        action_priority: Optional[ActionPriority],
        action_catalog: Optional[str],
        action_folder: Optional[str],
        storage_bucket_name: Optional[str],
        storage_bucket_directory_path: Optional[str],
        extraction_response: ExtractionResponse,
    ) -> StartOperationResponse:
        if project_type == ProjectType.PRETRAINED:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/extractors/{document_type_id}/validation/start"
            )
        else:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/{tag}/document-types/{document_type_id}/validation/start"
            )

        operation_id = (
            await self.request_async(
                "POST",
                url=url,
                params={"api-version": 1.1},
                headers=self._get_common_headers(),
                json={
                    "extractionResult": extraction_response.extraction_result.model_dump(),
                    "documentId": extraction_response.extraction_result.document_id,
                    "actionTitle": action_title,
                    "actionPriority": action_priority,
                    "actionCatalog": action_catalog,
                    "actionFolder": action_folder,
                    "storageBucketName": storage_bucket_name,
                    "allowChangeOfDocumentType": True,
                    "storageBucketDirectoryPath": storage_bucket_directory_path,
                },
            )
        ).json()["operationId"]

        return StartOperationResponse(
            operation_id=operation_id,
            document_id=extraction_response.extraction_result.document_id,
            project_id=project_id,
            tag=tag,
        )

    @traced(name="documents_start_ixp_extraction_validation", run_type="uipath")
    def start_ixp_extraction_validation(
        self,
        extraction_response: ExtractionResponseIXP,
        action_title: str,
        action_catalog: Optional[str] = None,
        action_priority: Optional[ActionPriority] = None,
        action_folder: Optional[str] = None,
        storage_bucket_name: Optional[str] = None,
        storage_bucket_directory_path: Optional[str] = None,
    ) -> StartOperationResponse:
        """Start an IXP extraction validation action without waiting for results (non-blocking).

        Args:
            extraction_response (ExtractionResponseIXP): The extraction response from the IXP extraction process.
            action_title (str): The title of the validation action.
            action_catalog (str, optional): The catalog of the validation action.
            action_priority (ActionPriority, optional): The priority of the validation action.
            action_folder (str, optional): The folder of the validation action.
            storage_bucket_name (str, optional): The name of the storage bucket where validation data will be stored.
            storage_bucket_directory_path (str, optional): The directory path within the storage bucket.

        Returns:
            StartOperationResponse: Contains the operation_id, document_id, project_id, and tag.

        Examples:
            ```python
            start_operation_response = service.start_ixp_extraction_validation(
                action_title="Validate IXP Extraction",
                action_priority=ActionPriority.HIGH,
                action_catalog="DefaultCatalog",
                action_folder="Validations",
                storage_bucket_name="my-storage-bucket",
                storage_bucket_directory_path="validations/ixp",
                extraction_response=extraction_response,
            )
            # start_operation_response can be used to poll for validation results later
            ```
        """
        return self._start_extraction_validation(
            project_id=extraction_response.project_id,
            project_type=ProjectType.IXP,
            tag=extraction_response.tag,
            document_type_id=str(UUID(int=0)),
            action_title=action_title,
            action_priority=action_priority,
            action_catalog=action_catalog,
            action_folder=action_folder,
            storage_bucket_name=storage_bucket_name,
            storage_bucket_directory_path=storage_bucket_directory_path,
            extraction_response=extraction_response,
        )

    @traced(
        name="documents_start_ixp_extraction_validation_async",
        run_type="uipath",
    )
    async def start_ixp_extraction_validation_async(
        self,
        extraction_response: ExtractionResponseIXP,
        action_title: str,
        action_catalog: Optional[str] = None,
        action_priority: Optional[ActionPriority] = None,
        action_folder: Optional[str] = None,
        storage_bucket_name: Optional[str] = None,
        storage_bucket_directory_path: Optional[str] = None,
    ) -> StartOperationResponse:
        """Asynchronous version of the [`start_ixp_extraction_validation`][uipath.platform.documents._documents_service.DocumentsService.start_ixp_extraction_validation] method."""
        return await self._start_extraction_validation_async(
            project_id=extraction_response.project_id,
            project_type=ProjectType.IXP,
            tag=extraction_response.tag,
            document_type_id=str(UUID(int=0)),
            action_title=action_title,
            action_priority=action_priority,
            action_catalog=action_catalog,
            action_folder=action_folder,
            storage_bucket_name=storage_bucket_name,
            storage_bucket_directory_path=storage_bucket_directory_path,
            extraction_response=extraction_response,
        )

    @traced(
        name="documents_retrieve_ixp_extraction_validation_result",
        run_type="uipath",
    )
    def retrieve_ixp_extraction_validation_result(
        self,
        project_id: str,
        tag: str,
        operation_id: str,
    ) -> ValidateExtractionAction:
        """Retrieve the result of an IXP create validate extraction action operation (single-shot, non-blocking).

        This method retrieves the result of an IXP create validate extraction action that was previously started
        with `start_ixp_extraction_validation`. It does not poll - it makes a single request and
        returns the result if available, or raises an exception if not complete.

        Args:
            operation_id (str): The operation ID returned from `start_ixp_extraction_validation`.
            project_id (str): The ID of the IXP project.
            tag (str): The tag of the published project version.

        Returns:
            ValidateExtractionAction: The validation action

        Raises:
            OperationNotCompleteException: If the validation action is not yet complete.
            OperationFailedException: If the validation action has failed.

        Examples:
            ```python
            # After receiving a callback/webhook that validation is complete:
            validation_result = service.retrieve_ixp_extraction_validation_result(
                operation_id=start_operation_response.operation_id,
                project_id=start_operation_response.project_id,
                tag=start_operation_response.tag,
            )
            ```
        """
        document_type_id = str(UUID(int=0))

        url = Endpoint(
            f"/du_/api/framework/projects/{project_id}/{tag}/document-types/{document_type_id}/validation/result/{operation_id}"
        )

        result = self._retrieve_operation_result(
            url=url,
            operation_id=operation_id,
            operation_name="IXP Create Validate Extraction Action",
        )

        result["projectId"] = project_id
        result["projectType"] = ProjectType.IXP
        result["tag"] = tag
        result["documentTypeId"] = str(UUID(int=0))
        result["operationId"] = operation_id

        return ValidateExtractionAction.model_validate(result)

    @traced(
        name="documents_retrieve_ixp_extraction_validation_result_async",
        run_type="uipath",
    )
    async def retrieve_ixp_extraction_validation_result_async(
        self,
        project_id: str,
        tag: str,
        operation_id: str,
    ) -> ValidateExtractionAction:
        """Asynchronous version of the [`retrieve_ixp_extraction_validation_result`][uipath.platform.documents._documents_service.DocumentsService.retrieve_ixp_extraction_validation_result] method."""
        document_type_id = str(UUID(int=0))

        url = Endpoint(
            f"/du_/api/framework/projects/{project_id}/{tag}/document-types/{document_type_id}/validation/result/{operation_id}"
        )

        result = await self._retrieve_operation_result_async(
            url=url,
            operation_id=operation_id,
            operation_name="IXP Create Validate Extraction Action",
        )

        result["projectId"] = project_id
        result["projectType"] = ProjectType.IXP
        result["tag"] = tag
        result["documentTypeId"] = str(UUID(int=0))
        result["operationId"] = operation_id

        return ValidateExtractionAction.model_validate(result)

    def _get_classification_validation_result(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        operation_id: str,
    ) -> Dict:
        if project_type == ProjectType.PRETRAINED:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/classifiers/ml-classification/validation/result/{operation_id}"
            )
        else:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/{tag}/classifiers/validation/result/{operation_id}"
            )

        return self.request(
            method="GET",
            url=url,
            params={"api-version": 1.1},
            headers=self._get_common_headers(),
        ).json()

    async def _get_classification_validation_result_async(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        operation_id: str,
    ) -> Dict:
        if project_type == ProjectType.PRETRAINED:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/classifiers/ml-classification/validation/result/{operation_id}"
            )
        else:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/{tag}/classifiers/validation/result/{operation_id}"
            )

        return (
            await self.request_async(
                method="GET",
                url=url,
                params={"api-version": 1.1},
                headers=self._get_common_headers(),
            )
        ).json()

    def _get_extraction_validation_result(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        document_type_id: str,
        operation_id: str,
    ) -> Dict:
        if project_type == ProjectType.PRETRAINED:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/extractors/{document_type_id}/validation/result/{operation_id}"
            )
        else:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/{tag}/document-types/{document_type_id}/validation/result/{operation_id}"
            )

        return self.request(
            method="GET",
            url=url,
            params={"api-version": 1.1},
            headers=self._get_common_headers(),
        ).json()

    async def _get_extraction_validation_result_async(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        document_type_id: str,
        operation_id: str,
    ) -> Dict:
        if project_type == ProjectType.PRETRAINED:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/extractors/{document_type_id}/validation/result/{operation_id}"
            )
        else:
            url = Endpoint(
                f"/du_/api/framework/projects/{project_id}/{tag}/document-types/{document_type_id}/validation/result/{operation_id}"
            )

        return (
            await self.request_async(
                method="GET",
                url=url,
                params={"api-version": 1.1},
                headers=self._get_common_headers(),
            )
        ).json()

    def _wait_for_create_validate_classification_action(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        operation_id: str,
    ) -> ValidateClassificationAction:
        def result_getter() -> Tuple[Any, Optional[Any], Optional[Any]]:
            result = self._get_classification_validation_result(
                project_id=project_id,
                project_type=project_type,
                tag=tag,
                operation_id=operation_id,
            )
            return (
                result["status"],
                result.get("error", None),
                result.get("result", None),
            )

        response = self._wait_for_operation(
            result_getter=result_getter,
            wait_statuses=["NotStarted", "Running"],
            success_status="Succeeded",
        )

        response["projectId"] = project_id
        response["projectType"] = project_type
        response["tag"] = tag
        response["operationId"] = operation_id
        return ValidateClassificationAction.model_validate(response)

    async def _wait_for_create_validate_classification_action_async(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        operation_id: str,
    ) -> ValidateClassificationAction:
        async def result_getter() -> Tuple[Any, Optional[Any], Optional[Any]]:
            result = await self._get_classification_validation_result_async(
                project_id=project_id,
                project_type=project_type,
                tag=tag,
                operation_id=operation_id,
            )
            return (
                result["status"],
                result.get("error", None),
                result.get("result", None),
            )

        response = await self._wait_for_operation_async(
            result_getter=result_getter,
            wait_statuses=["NotStarted", "Running"],
            success_status="Succeeded",
        )

        response["projectId"] = project_id
        response["projectType"] = project_type
        response["tag"] = tag
        response["operationId"] = operation_id
        return ValidateClassificationAction.model_validate(response)

    def _wait_for_create_validate_extraction_action(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        document_type_id: str,
        operation_id: str,
    ) -> ValidateExtractionAction:
        def result_getter() -> Tuple[Any, Optional[Any], Optional[Any]]:
            result = self._get_extraction_validation_result(
                project_id=project_id,
                project_type=project_type,
                tag=tag,
                document_type_id=document_type_id,
                operation_id=operation_id,
            )
            return (
                result["status"],
                result.get("error", None),
                result.get("result", None),
            )

        response = self._wait_for_operation(
            result_getter=result_getter,
            wait_statuses=["NotStarted", "Running"],
            success_status="Succeeded",
        )

        response["projectId"] = project_id
        response["projectType"] = project_type
        response["tag"] = tag
        response["documentTypeId"] = document_type_id
        response["operationId"] = operation_id
        return ValidateExtractionAction.model_validate(response)

    async def _wait_for_create_validate_extraction_action_async(
        self,
        project_id: str,
        project_type: ProjectType,
        tag: Optional[str],
        document_type_id: str,
        operation_id: str,
    ) -> ValidateExtractionAction:
        async def result_getter_async() -> Tuple[Any, Optional[Any], Optional[Any]]:
            result = await self._get_extraction_validation_result_async(
                project_id=project_id,
                project_type=project_type,
                tag=tag,
                document_type_id=document_type_id,
                operation_id=operation_id,
            )
            return (
                result["status"],
                result.get("error", None),
                result.get("result", None),
            )

        response = await self._wait_for_operation_async(
            result_getter=result_getter_async,
            wait_statuses=["NotStarted", "Running"],
            success_status="Succeeded",
        )

        response["projectId"] = project_id
        response["projectType"] = project_type
        response["tag"] = tag
        response["documentTypeId"] = document_type_id
        response["operationId"] = operation_id
        return ValidateExtractionAction.model_validate(response)

    @traced(name="documents_create_validate_classification_action", run_type="uipath")
    def create_validate_classification_action(
        self,
        classification_results: List[ClassificationResult],
        action_title: str,
        action_priority: Optional[ActionPriority] = None,
        action_catalog: Optional[str] = None,
        action_folder: Optional[str] = None,
        storage_bucket_name: Optional[str] = None,
        storage_bucket_directory_path: Optional[str] = None,
    ) -> ValidateClassificationAction:
        """Create a validate classification action for a document based on the classification results. More details about validation actions can be found in the [official documentation](https://docs.uipath.com/ixp/automation-cloud/latest/user-guide/validating-classifications).

        Args:
            classification_results (List[ClassificationResult]): The classification results to be validated, typically obtained from the [`classify`][uipath.platform.documents._documents_service.DocumentsService.classify] method.
            action_title (str): Title of the action.
            action_priority (ActionPriority, optional): Priority of the action.
            action_catalog (str, optional): Catalog of the action.
            action_folder (str, optional): Folder of the action.
            storage_bucket_name (str, optional): Name of the storage bucket.
            storage_bucket_directory_path (str, optional): Directory path in the storage bucket.

        Returns:
            ValidateClassificationAction: The created validate classification action.

        Examples:
            ```python
            validation_action = service.create_validate_classification_action(
                action_title="Test Validation Action",
                action_priority=ActionPriority.MEDIUM,
                action_catalog="default_du_actions",
                action_folder="Shared",
                storage_bucket_name="du_storage_bucket",
                storage_bucket_directory_path="TestDirectory",
                classification_results=classification_results,
            )
            ```
        """
        if not classification_results:
            raise ValueError("`classification_results` must not be empty")

        operation_id = self._start_classification_validation(
            project_id=classification_results[0].project_id,
            project_type=classification_results[0].project_type,
            tag=classification_results[0].tag,
            action_title=action_title,
            action_priority=action_priority,
            action_catalog=action_catalog,
            action_folder=action_folder,
            storage_bucket_name=storage_bucket_name,
            storage_bucket_directory_path=storage_bucket_directory_path,
            classification_results=classification_results,
        )

        return self._wait_for_create_validate_classification_action(
            project_id=classification_results[0].project_id,
            project_type=classification_results[0].project_type,
            tag=classification_results[0].tag,
            operation_id=operation_id,
        )

    @traced(name="documents_create_validate_classification_action", run_type="uipath")
    async def create_validate_classification_action_async(
        self,
        classification_results: List[ClassificationResult],
        action_title: str,
        action_priority: Optional[ActionPriority] = None,
        action_catalog: Optional[str] = None,
        action_folder: Optional[str] = None,
        storage_bucket_name: Optional[str] = None,
        storage_bucket_directory_path: Optional[str] = None,
    ) -> ValidateClassificationAction:
        """Asynchronous version of the [`create_validation_action`][uipath.platform.documents._documents_service.DocumentsService.create_validate_classification_action] method."""
        if not classification_results:
            raise ValueError("`classification_results` must not be empty")

        operation_id = await self._start_classification_validation_async(
            project_id=classification_results[0].project_id,
            project_type=classification_results[0].project_type,
            tag=classification_results[0].tag,
            action_title=action_title,
            action_priority=action_priority,
            action_catalog=action_catalog,
            action_folder=action_folder,
            storage_bucket_name=storage_bucket_name,
            storage_bucket_directory_path=storage_bucket_directory_path,
            classification_results=classification_results,
        )

        return await self._wait_for_create_validate_classification_action_async(
            project_id=classification_results[0].project_id,
            project_type=classification_results[0].project_type,
            tag=classification_results[0].tag,
            operation_id=operation_id,
        )

    @traced(name="documents_create_validate_extraction_action", run_type="uipath")
    def create_validate_extraction_action(
        self,
        extraction_response: ExtractionResponse,
        action_title: str,
        action_priority: Optional[ActionPriority] = None,
        action_catalog: Optional[str] = None,
        action_folder: Optional[str] = None,
        storage_bucket_name: Optional[str] = None,
        storage_bucket_directory_path: Optional[str] = None,
    ) -> ValidateExtractionAction:
        """Create a validate extraction action for a document based on the extraction response. More details about validation actions can be found in the [official documentation](https://docs.uipath.com/ixp/automation-cloud/latest/user-guide/validating-extractions).

        Args:
            extraction_response (ExtractionResponse): The extraction result to be validated, typically obtained from the [`extract`][uipath.platform.documents._documents_service.DocumentsService.extract] method.
            action_title (str): Title of the action.
            action_priority (ActionPriority, optional): Priority of the action.
            action_catalog (str, optional): Catalog of the action.
            action_folder (str, optional): Folder of the action.
            storage_bucket_name (str, optional): Name of the storage bucket.
            storage_bucket_directory_path (str, optional): Directory path in the storage bucket.

        Returns:
            ValidateClassificationAction: The created validation action.

        Examples:
            ```python
            validation_action = service.create_validate_extraction_action(
                action_title="Test Validation Action",
                action_priority=ActionPriority.MEDIUM,
                action_catalog="default_du_actions",
                action_folder="Shared",
                storage_bucket_name="du_storage_bucket",
                storage_bucket_directory_path="TestDirectory",
                extraction_response=extraction_response,
            )
            ```
        """
        operation_id = self._start_extraction_validation(
            project_id=extraction_response.project_id,
            project_type=extraction_response.project_type,
            tag=extraction_response.tag,
            document_type_id=extraction_response.document_type_id,
            action_title=action_title,
            action_priority=action_priority,
            action_catalog=action_catalog,
            action_folder=action_folder,
            storage_bucket_name=storage_bucket_name,
            storage_bucket_directory_path=storage_bucket_directory_path,
            extraction_response=extraction_response,
        ).operation_id

        return self._wait_for_create_validate_extraction_action(
            project_id=extraction_response.project_id,
            project_type=extraction_response.project_type,
            tag=extraction_response.tag,
            document_type_id=extraction_response.document_type_id,
            operation_id=operation_id,
        )

    @traced(name="documents_create_validate_extraction_action_async", run_type="uipath")
    async def create_validate_extraction_action_async(
        self,
        extraction_response: ExtractionResponse,
        action_title: str,
        action_priority: Optional[ActionPriority] = None,
        action_catalog: Optional[str] = None,
        action_folder: Optional[str] = None,
        storage_bucket_name: Optional[str] = None,
        storage_bucket_directory_path: Optional[str] = None,
    ) -> ValidateExtractionAction:
        """Asynchronous version of the [`create_validation_action`][uipath.platform.documents._documents_service.DocumentsService.create_validate_extraction_action] method."""
        operation_id = (
            await self._start_extraction_validation_async(
                project_id=extraction_response.project_id,
                project_type=extraction_response.project_type,
                tag=extraction_response.tag,
                document_type_id=extraction_response.document_type_id,
                action_title=action_title,
                action_priority=action_priority,
                action_catalog=action_catalog,
                action_folder=action_folder,
                storage_bucket_name=storage_bucket_name,
                storage_bucket_directory_path=storage_bucket_directory_path,
                extraction_response=extraction_response,
            )
        ).operation_id

        return await self._wait_for_create_validate_extraction_action_async(
            project_id=extraction_response.project_id,
            project_type=extraction_response.project_type,
            tag=extraction_response.tag,
            document_type_id=extraction_response.document_type_id,
            operation_id=operation_id,
        )

    @traced(name="documents_get_validate_classification_result", run_type="uipath")
    def get_validate_classification_result(
        self, validation_action: ValidateClassificationAction
    ) -> List[ClassificationResult]:
        """Get the result of a validate classification action.

        Note:
            This method will block until the validation action is completed, meaning the user has completed the validation in UiPath Action Center.

        Args:
            validation_action (ValidateClassificationAction): The validation action to get the result for, typically obtained from the [`create_validate_classification_action`][uipath.platform.documents._documents_service.DocumentsService.create_validate_classification_action] method.

        Returns:
            List[ClassificationResult]: The validated classification results.

        Examples:
            ```python
            validated_results = service.get_validate_classification_result(validate_classification_action)
            ```
        """

        def result_getter() -> Tuple[str, None, Any]:
            result = self._get_classification_validation_result(
                project_id=validation_action.project_id,
                project_type=validation_action.project_type,
                tag=validation_action.tag,
                operation_id=validation_action.operation_id,
            )
            return (result["result"]["actionStatus"], None, result["result"])

        response = self._wait_for_operation(
            result_getter=result_getter,
            wait_statuses=["Unassigned", "Pending"],
            success_status="Completed",
        )
        classification_results = []
        for cr in response["validatedClassificationResults"]:
            cr["ProjectId"] = validation_action.project_id
            cr["ProjectType"] = validation_action.project_type
            cr["Tag"] = validation_action.tag
            classification_results.append(ClassificationResult.model_validate(cr))

        return classification_results

    @traced(
        name="documents_get_validate_classification_result_async", run_type="uipath"
    )
    async def get_validate_classification_result_async(
        self, validation_action: ValidateClassificationAction
    ) -> List[ClassificationResult]:
        """Asynchronous version of the [`get_validation_result`][uipath.platform.documents._documents_service.DocumentsService.get_validate_classification_result] method."""

        async def result_getter() -> Tuple[str, None, Any]:
            result = await self._get_classification_validation_result_async(
                project_id=validation_action.project_id,
                project_type=validation_action.project_type,
                tag=validation_action.tag,
                operation_id=validation_action.operation_id,
            )
            return (result["result"]["actionStatus"], None, result["result"])

        response = await self._wait_for_operation_async(
            result_getter=result_getter,
            wait_statuses=["Unassigned", "Pending"],
            success_status="Completed",
        )
        classification_results = []
        for cr in response["validatedClassificationResults"]:
            cr["ProjectId"] = validation_action.project_id
            cr["ProjectType"] = validation_action.project_type
            cr["Tag"] = validation_action.tag
            classification_results.append(ClassificationResult.model_validate(cr))

        return classification_results

    @traced(name="documents_get_validate_extraction_result", run_type="uipath")
    def get_validate_extraction_result(
        self, validation_action: ValidateExtractionAction
    ) -> Union[ExtractionResponse, ExtractionResponseIXP]:
        """Get the result of a validate extraction action.

        Note:
            This method will block until the validation action is completed, meaning the user has completed the validation in UiPath Action Center.

        Args:
            validation_action (ValidateClassificationAction): The validation action to get the result for, typically obtained from the [`create_validate_extraction_action`][uipath.platform.documents._documents_service.DocumentsService.create_validate_extraction_action] method.

        Returns:
            Union[ExtractionResponse, ExtractionResponseIXP]: The validated extraction response.

        Examples:
            ```python
            validated_result = service.get_validate_extraction_result(validate_extraction_action)
            ```
        """

        def result_getter() -> Tuple[str, None, Any]:
            result = self._get_extraction_validation_result(
                project_id=validation_action.project_id,
                project_type=validation_action.project_type,
                tag=validation_action.tag,
                document_type_id=validation_action.document_type_id,
                operation_id=validation_action.operation_id,
            )
            return (result["result"]["actionStatus"], None, result["result"])

        response = self._wait_for_operation(
            result_getter=result_getter,
            wait_statuses=["Unassigned", "Pending"],
            success_status="Completed",
        )
        response["extractionResult"] = response.pop("validatedExtractionResults")
        response["projectId"] = validation_action.project_id
        response["tag"] = validation_action.tag
        response["documentTypeId"] = validation_action.document_type_id
        response["projectType"] = validation_action.project_type

        if validation_action.project_type == ProjectType.IXP:
            return ExtractionResponseIXP.model_validate(response)

        return ExtractionResponse.model_validate(response)

    @traced(name="documents_get_validate_extraction_result_async", run_type="uipath")
    async def get_validate_extraction_result_async(
        self, validation_action: ValidateExtractionAction
    ) -> Union[ExtractionResponse, ExtractionResponseIXP]:
        """Asynchronous version of the [`get_validation_result`][uipath.platform.documents._documents_service.DocumentsService.get_validate_extraction_result] method."""

        async def result_getter() -> Tuple[str, None, Any]:
            result = await self._get_extraction_validation_result_async(
                project_id=validation_action.project_id,
                project_type=validation_action.project_type,
                tag=validation_action.tag,
                document_type_id=validation_action.document_type_id,
                operation_id=validation_action.operation_id,
            )
            return (result["result"]["actionStatus"], None, result["result"])

        response = await self._wait_for_operation_async(
            result_getter=result_getter,
            wait_statuses=["Unassigned", "Pending"],
            success_status="Completed",
        )
        response["extractionResult"] = response.pop("validatedExtractionResults")
        response["projectId"] = validation_action.project_id
        response["tag"] = validation_action.tag
        response["documentTypeId"] = validation_action.document_type_id
        response["projectType"] = validation_action.project_type

        if validation_action.project_type == ProjectType.IXP:
            return ExtractionResponseIXP.model_validate(response)

        return ExtractionResponse.model_validate(response)



"""Protocol definition for CRUD services.

This module defines the CRUDServiceProtocol that services must implement
to be compatible with the CLI command generator.

The protocol uses Python's Protocol (PEP 544) for structural subtyping,
allowing duck-typing without requiring explicit inheritance.

Example:
    >>> from typing import Protocol, runtime_checkable
    >>> from ._service_protocol import CRUDServiceProtocol
    >>>
    >>> # A service that implements the protocol
    >>> class BucketsService:
    ...     def list(self, folder_path=None, folder_key=None):
    ...         return iter([])
    ...
    ...     def retrieve(self, name, folder_path=None, folder_key=None):
    ...         return {"name": name}
    ...
    ...     # ... other CRUD methods
    >>>
    >>> # No explicit inheritance needed - structural typing
    >>> def use_service(service: CRUDServiceProtocol):
    ...     return list(service.list())
"""

from typing import Any, Iterator, Protocol, runtime_checkable


@runtime_checkable
class CRUDServiceProtocol(Protocol):
    """Protocol for services that support standard CRUD operations.

    Services implementing this protocol can use the ServiceCLIGenerator
    to automatically generate CLI commands.

    All methods support folder-scoped operations via folder_path or folder_key
    parameters. This matches the UiPath platform's hierarchical folder structure.

    Required Methods:
        list: List all resources (returns iterator)
        retrieve: Get a single resource by identifier
        create: Create a new resource
        delete: Delete a resource by identifier

    Optional Methods (NOT in protocol, checked separately):
        exists: Check if a resource exists (will be used if present on service)

    Note:
        This is a Protocol, not a base class. Services don't need to inherit
        from it; they just need to implement the required methods with
        compatible signatures.

        The exists() method is NOT part of the protocol to allow optional
        implementation. Use hasattr(service, 'exists') to check for it.
    """

    def list(
        self,
        *,
        folder_path: str | None = None,
        folder_key: str | None = None,
        **kwargs: Any,
    ) -> Iterator[Any]:
        """List all resources in the specified folder.

        Args:
            folder_path: Folder path (e.g., "Shared")
            folder_key: Folder UUID key
            **kwargs: Additional service-specific parameters

        Returns:
            Iterator of resource objects

        Example:
            >>> buckets_service.list(folder_path="Shared")
            <iterator of Bucket objects>
        """
        ...

    def retrieve(
        self,
        name: str,
        *,
        folder_path: str | None = None,
        folder_key: str | None = None,
        **kwargs: Any,
    ) -> Any:
        """Retrieve a single resource by identifier.

        Args:
            name: Resource identifier (usually name, but could be key)
            folder_path: Folder path (e.g., "Shared")
            folder_key: Folder UUID key
            **kwargs: Additional service-specific parameters

        Returns:
            Resource object

        Raises:
            LookupError: If resource not found

        Example:
            >>> buckets_service.retrieve("my-bucket", folder_path="Shared")
            Bucket(name="my-bucket", ...)
        """
        ...

    def create(
        self,
        name: str,
        *,
        folder_path: str | None = None,
        folder_key: str | None = None,
        **kwargs: Any,
    ) -> Any:
        """Create a new resource.

        Args:
            name: Resource name
            folder_path: Folder path (e.g., "Shared")
            folder_key: Folder UUID key
            **kwargs: Additional service-specific parameters (from metadata)

        Returns:
            Created resource object

        Example:
            >>> buckets_service.create(
            ...     "my-bucket",
            ...     folder_path="Shared",
            ...     description="My bucket"
            ... )
            Bucket(name="my-bucket", description="My bucket", ...)
        """
        ...

    def delete(
        self,
        name: str,
        *,
        folder_path: str | None = None,
        folder_key: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Delete a resource by identifier.

        Args:
            name: Resource identifier to delete
            folder_path: Folder path (e.g., "Shared")
            folder_key: Folder UUID key
            **kwargs: Additional service-specific parameters

        Raises:
            LookupError: If resource not found

        Example:
            >>> buckets_service.delete("my-bucket", folder_path="Shared")
        """
        ...


def has_exists_method(service: Any) -> bool:
    """Check if a service has the optional exists() method.

    This is separate from the Protocol because exists() is optional.
    Not all services need to implement it.

    Args:
        service: Service instance to check

    Returns:
        True if service has a callable exists() method

    Example:
        >>> class ServiceWithExists:
        ...     def exists(self, name, **kwargs):
        ...         return True
        >>>
        >>> service = ServiceWithExists()
        >>> has_exists_method(service)
        True
    """
    return hasattr(service, "exists") and callable(service.exists)


def validate_service_protocol(service: Any, service_name: str) -> None:
    """Validate that a service implements the CRUDServiceProtocol.

    Args:
        service: Service instance to validate
        service_name: Name of the service for error messages

    Raises:
        TypeError: If service doesn't implement required methods

    Example:
        >>> class IncompleteService:
        ...     def list(self):
        ...         pass
        ...     # Missing other methods
        >>>
        >>> validate_service_protocol(IncompleteService(), "incomplete")
        Traceback (most recent call last):
        ...
        TypeError: Service 'incomplete' must implement: retrieve, create, delete
    """
    required_methods = ["list", "retrieve", "create", "delete"]
    missing_methods = []
    for method_name in required_methods:
        if not hasattr(service, method_name) or not callable(
            getattr(service, method_name)
        ):
            missing_methods.append(method_name)

    if missing_methods:
        raise TypeError(
            f"Service '{service_name}' must implement: {', '.join(missing_methods)}"
        )


__all__ = ["CRUDServiceProtocol", "validate_service_protocol", "has_exists_method"]



"""Service command groups for UiPath CLI.

All services are explicitly imported to ensure:
- File renames break loudly (import error) not silently
- Clear dependency graph
- Fast startup (no auto-discovery overhead)
- Debuggable registration
"""

from .cli_assets import assets
from .cli_buckets import buckets

__all__ = ["assets", "buckets", "register_service_commands"]


def register_service_commands(cli_group):
    """Register all service command groups with the root CLI.

    This function maintains explicitness while reducing registration boilerplate.
    Benefits:
    - File renames break loudly (import error) not silently
    - Clear list of all registered services
    - Easy to comment out services during development

    Args:
        cli_group: The root Click group to register services with

    Returns:
        The cli_group for method chaining

    Industry Precedent:
        AWS CLI, Azure CLI, and gcloud all use explicit registration.
    """
    services = [assets, buckets]

    for service in services:
        cli_group.add_command(service)

    return cli_group



"""Tests for UiPathEvalRuntime suspend/resume event publishing.

This module tests:
- Normal flow: CREATE_EVAL_RUN event is published, eval_run_id saved to disk
- Resume flow: CREATE_EVAL_RUN event is NOT published, eval_run_id loaded from disk
- UPDATE_EVAL_RUN works correctly after evaluators complete
- Ensures no duplicate eval run entries in StudioWeb
"""

from pathlib import Path
from typing import Any, AsyncGenerator
from unittest.mock import AsyncMock

import pytest
from uipath.core.tracing import UiPathTraceManager
from uipath.runtime import (
    UiPathExecuteOptions,
    UiPathResumableStorageProtocol,
    UiPathRuntimeEvent,
    UiPathRuntimeFactorySettings,
    UiPathRuntimeProtocol,
    UiPathRuntimeResult,
    UiPathRuntimeStatus,
    UiPathStreamOptions,
)
from uipath.runtime.schema import UiPathRuntimeSchema

from uipath._cli._evals._evaluate import evaluate
from uipath._cli._evals._runtime import UiPathEvalContext
from uipath._events._event_bus import EventBus
from uipath._events._events import EvaluationEvents


class MockRuntimeSchema(UiPathRuntimeSchema):
    """Mock schema for testing."""

    def __init__(self):
        super().__init__(
            filePath="test.py",
            uniqueId="test",
            type="agent",
            input={"type": "object", "properties": {}},
            output={"type": "object", "properties": {}},
        )


class SuspendingRuntime:
    """Test runtime that returns SUSPENDED status."""

    async def execute(
        self,
        input: dict[str, Any] | None = None,
        options: UiPathExecuteOptions | None = None,
    ) -> UiPathRuntimeResult:
        # Simulate an agent that suspends
        # Returning SUSPENDED status is sufficient for our test
        return UiPathRuntimeResult(
            output={},
            status=UiPathRuntimeStatus.SUSPENDED,
        )

    async def stream(
        self,
        input: dict[str, Any] | None = None,
        options: UiPathStreamOptions | None = None,
    ) -> AsyncGenerator[UiPathRuntimeEvent, None]:
        yield UiPathRuntimeResult(
            output={},
            status=UiPathRuntimeStatus.SUSPENDED,
        )

    async def get_schema(self) -> UiPathRuntimeSchema:
        return MockRuntimeSchema()

    async def dispose(self) -> None:
        pass


class SuccessfulRuntime:
    """Test runtime that returns SUCCESSFUL status."""

    async def execute(
        self,
        input: dict[str, Any] | None = None,
        options: UiPathExecuteOptions | None = None,
    ) -> UiPathRuntimeResult:
        return UiPathRuntimeResult(
            output={"result": "success"},
            status=UiPathRuntimeStatus.SUCCESSFUL,
        )

    async def stream(
        self,
        input: dict[str, Any] | None = None,
        options: UiPathStreamOptions | None = None,
    ) -> AsyncGenerator[UiPathRuntimeEvent, None]:
        yield UiPathRuntimeResult(
            output={"result": "success"},
            status=UiPathRuntimeStatus.SUCCESSFUL,
        )

    async def get_schema(self) -> UiPathRuntimeSchema:
        return MockRuntimeSchema()

    async def dispose(self) -> None:
        pass


class MockFactory:
    """Mock factory for creating test runtimes."""

    def __init__(self, runtime_creator):
        self.runtime_creator = runtime_creator

    def discover_entrypoints(self) -> list[str]:
        return ["test"]

    async def get_storage(self) -> UiPathResumableStorageProtocol | None:
        return None

    async def get_settings(self) -> UiPathRuntimeFactorySettings | None:
        return None

    async def new_runtime(
        self, entrypoint: str, runtime_id: str, **kwargs
    ) -> UiPathRuntimeProtocol:
        return await self.runtime_creator()

    async def dispose(self) -> None:
        pass


@pytest.fixture
def context():
    """Create eval context."""
    context = UiPathEvalContext()
    context.eval_set = str(
        Path(__file__).parent / "evals" / "eval-sets" / "default.json"
    )
    return context


@pytest.fixture
def event_bus():
    """Create event bus with mocked publish method."""
    bus = EventBus()
    bus.publish = AsyncMock()  # type: ignore[method-assign]
    return bus


@pytest.fixture
def trace_manager():
    """Create trace manager."""
    return UiPathTraceManager()


class TestNormalFlowCreateEvalRun:
    """Tests for normal execution flow - CREATE_EVAL_RUN should be published."""

    async def test_publishes_create_eval_run_on_successful_execution(
        self, context, event_bus, trace_manager
    ):
        """Test that CREATE_EVAL_RUN is published during normal successful execution."""

        # Arrange
        async def create_runtime():
            return SuccessfulRuntime()

        factory = MockFactory(create_runtime)

        # Act
        await evaluate(factory, trace_manager, context, event_bus)

        # Assert
        # Check that CREATE_EVAL_RUN was published
        create_calls = [
            call
            for call in event_bus.publish.call_args_list
            if call[0][0] == EvaluationEvents.CREATE_EVAL_RUN
        ]
        assert len(create_calls) > 0, (
            "CREATE_EVAL_RUN should be published in normal flow"
        )

    async def test_publishes_create_eval_run_on_suspend(
        self, context, event_bus, trace_manager
    ):
        """Test that CREATE_EVAL_RUN is published when agent suspends."""

        # Arrange
        async def create_runtime():
            return SuspendingRuntime()

        factory = MockFactory(create_runtime)

        # Act
        await evaluate(factory, trace_manager, context, event_bus)

        # Assert
        # Check that CREATE_EVAL_RUN was published
        create_calls = [
            call
            for call in event_bus.publish.call_args_list
            if call[0][0] == EvaluationEvents.CREATE_EVAL_RUN
        ]
        assert len(create_calls) > 0, (
            "CREATE_EVAL_RUN should be published on suspend (initial execution)"
        )


class TestResumeFlowSkipsCreateEvalRun:
    """Tests for resume flow - CREATE_EVAL_RUN should NOT be published, eval_run_id loaded from disk."""

    async def test_skips_create_eval_run_on_resume(
        self, context, event_bus, trace_manager
    ):
        """Test that CREATE_EVAL_RUN is NOT published when resuming from checkpoint.

        During resume, the eval_run_id mapping is loaded from the persisted file
        (__uipath/eval_run_ids.json) instead of making a new CREATE_EVAL_RUN API call.
        This prevents duplicate entries in StudioWeb.
        """
        # Arrange
        context.resume = True  # Set resume flag

        async def create_runtime():
            return SuccessfulRuntime()

        factory = MockFactory(create_runtime)

        # Act
        await evaluate(factory, trace_manager, context, event_bus)

        # Assert
        # Check that CREATE_EVAL_RUN was NOT published
        create_calls = [
            call
            for call in event_bus.publish.call_args_list
            if call[0][0] == EvaluationEvents.CREATE_EVAL_RUN
        ]
        assert len(create_calls) == 0, (
            "CREATE_EVAL_RUN should NOT be published on resume (loaded from disk instead)"
        )

    async def test_publishes_update_eval_run_on_resume(
        self, context, event_bus, trace_manager
    ):
        """Test that UPDATE_EVAL_RUN is still published when resuming."""
        # Arrange
        context.resume = True  # Set resume flag

        async def create_runtime():
            return SuccessfulRuntime()

        factory = MockFactory(create_runtime)

        # Act
        await evaluate(factory, trace_manager, context, event_bus)

        # Assert
        # Check that UPDATE_EVAL_RUN was published
        update_calls = [
            call
            for call in event_bus.publish.call_args_list
            if call[0][0] == EvaluationEvents.UPDATE_EVAL_RUN
        ]
        assert len(update_calls) > 0, (
            "UPDATE_EVAL_RUN should still be published on resume"
        )

    async def test_resume_flag_false_by_default(self, context):
        """Test that resume flag is False by default in context."""
        assert context.resume is False, "resume should default to False"

    async def test_no_duplicate_entries_on_resume(
        self, context, event_bus, trace_manager
    ):
        """Test that resuming doesn't create duplicate entries (integration test concept).

        This test verifies that when resume=True:
        1. CREATE_EVAL_RUN is NOT called (prevents duplicate backend entries)
        2. Only UPDATE_EVAL_RUN is called (updates existing entry)

        The eval_run_id mapping is loaded from persisted state on disk.
        """
        # Arrange
        context.resume = True

        async def create_runtime():
            return SuccessfulRuntime()

        factory = MockFactory(create_runtime)

        # Act
        await evaluate(factory, trace_manager, context, event_bus)

        # Assert
        create_calls = [
            call
            for call in event_bus.publish.call_args_list
            if call[0][0] == EvaluationEvents.CREATE_EVAL_RUN
        ]
        update_calls = [
            call
            for call in event_bus.publish.call_args_list
            if call[0][0] == EvaluationEvents.UPDATE_EVAL_RUN
        ]

        assert len(create_calls) == 0, "Should NOT create new entry on resume"
        assert len(update_calls) > 0, "Should update existing entry on resume"


class TestSuspendResumeLifecycle:
    """Tests for complete suspend/resume lifecycle."""

    async def test_suspend_then_resume_lifecycle(self, context, trace_manager):
        """Test complete lifecycle: suspend (creates entry) then resume (updates entry)."""
        # Phase 1: Suspend
        event_bus_suspend = EventBus()
        event_bus_suspend.publish = AsyncMock()  # type: ignore[method-assign]

        async def create_suspending_runtime():
            return SuspendingRuntime()

        factory_suspend = MockFactory(create_suspending_runtime)

        await evaluate(factory_suspend, trace_manager, context, event_bus_suspend)

        # Assert suspend phase
        suspend_create_calls = [
            call
            for call in event_bus_suspend.publish.call_args_list
            if call[0][0] == EvaluationEvents.CREATE_EVAL_RUN
        ]
        suspend_update_calls = [
            call
            for call in event_bus_suspend.publish.call_args_list
            if call[0][0] == EvaluationEvents.UPDATE_EVAL_RUN
        ]

        assert len(suspend_create_calls) > 0, "Suspend phase should CREATE entry"
        assert len(suspend_update_calls) == 0, (
            "Suspend phase should NOT UPDATE - evalRun stays IN_PROGRESS until resume"
        )

        # Phase 2: Resume
        context.resume = True
        event_bus_resume = EventBus()
        event_bus_resume.publish = AsyncMock()  # type: ignore[method-assign]

        async def create_successful_runtime():
            return SuccessfulRuntime()

        factory_resume = MockFactory(create_successful_runtime)

        await evaluate(factory_resume, trace_manager, context, event_bus_resume)

        # Assert resume phase
        resume_create_calls = [
            call
            for call in event_bus_resume.publish.call_args_list
            if call[0][0] == EvaluationEvents.CREATE_EVAL_RUN
        ]
        resume_update_calls = [
            call
            for call in event_bus_resume.publish.call_args_list
            if call[0][0] == EvaluationEvents.UPDATE_EVAL_RUN
        ]

        assert len(resume_create_calls) == 0, "Resume phase should NOT CREATE new entry"
        assert len(resume_update_calls) > 0, "Resume phase should UPDATE existing entry"



"""Trajectory evaluator for analyzing execution paths and decision sequences."""

from typing import Any, Optional

from opentelemetry.sdk.trace import ReadableSpan
from pydantic import field_validator

from uipath.eval.models import EvaluationResult

from ..._utils.constants import COMMUNITY_agents_SUFFIX
from ...platform.chat import UiPathLlmChatService
from ...platform.chat.llm_gateway import RequiredToolChoice
from ..models.models import (
    AgentExecution,
    LLMResponse,
    NumericEvaluationResult,
    TrajectoryEvaluationTrace,
)
from .legacy_base_evaluator import (
    LegacyBaseEvaluator,
    LegacyEvaluationCriteria,
    LegacyEvaluatorConfig,
)
from .legacy_llm_helpers import create_evaluation_tool, extract_tool_call_response


class LegacyTrajectoryEvaluatorConfig(LegacyEvaluatorConfig):
    """Configuration for legacy trajectory evaluators."""

    name: str = "LegacyTrajectoryEvaluator"


class LegacyTrajectoryEvaluator(LegacyBaseEvaluator[LegacyTrajectoryEvaluatorConfig]):
    """Legacy evaluator that analyzes the trajectory/path taken to reach outputs."""

    prompt: str
    model: str
    expected_agent_behavior_placeholder: str = "{{ExpectedAgentBehavior}}"
    agent_run_history_placeholder: str = "{{AgentRunHistory}}"
    llm: Optional[UiPathLlmChatService] = None

    @field_validator("prompt")
    @classmethod
    def validate_prompt_placeholder(cls, v: str) -> str:
        """Validate that prompt contains required placeholders."""
        if "{{ExpectedAgentBehavior}}" not in v or "{{AgentRunHistory}}" not in v:
            raise ValueError(
                "Prompt must contain {ExpectedAgentBehavior} and {{AgentRunHistory}} placeholders"
            )
        return v

    def model_post_init(self, __context: Any):
        """Initialize the evaluator after model creation."""
        super().model_post_init(__context)

    def _initialize_llm(self):
        """Initialize the LLM used for evaluation."""
        from uipath.platform import UiPath

        uipath = UiPath()
        self.llm = uipath.llm

    async def evaluate(
        self,
        agent_execution: AgentExecution,
        evaluation_criteria: LegacyEvaluationCriteria,
    ) -> EvaluationResult:
        """Evaluate using trajectory analysis.

        Analyzes the execution path and decision sequence taken by the agent.

        Args:
            agent_execution: The execution details containing:
                - agent_input: The input received by the agent
                - actual_output: The actual output from the agent
                - agent_trace: The execution spans to use for the evaluation
                - expected_agent_behavior: The expected agent behavior
            evaluation_criteria: The criteria to evaluate
        Returns:
            EvaluationResult: Score based on trajectory analysis

        Raises:
            NotImplementedError: This evaluator is not yet implemented
        """
        # Lazily initialize the LLM on first evaluation call
        if self.llm is None:
            self._initialize_llm()

        evaluation_prompt = self._create_evaluation_prompt(
            expected_agent_behavior=agent_execution.expected_agent_behavior,
            agent_run_history=agent_execution.agent_trace,
        )
        llm_response = await self._get_llm_response(evaluation_prompt)

        return NumericEvaluationResult(
            score=llm_response.score,
            details=llm_response.justification,
        )

    def _create_evaluation_prompt(
        self,
        expected_agent_behavior: Any,
        agent_run_history: Any,
    ) -> str:
        """Create the evaluation prompt for the LLM."""
        formatted_prompt = self.prompt.replace(
            self.expected_agent_behavior_placeholder,
            str(expected_agent_behavior),
        )

        # Trim extra properties from the spans (such as timestamps which are not relevant to the eval)
        if (
            isinstance(agent_run_history, list)
            and agent_run_history
            and isinstance(agent_run_history[0], ReadableSpan)
        ):
            trajectory_trace = TrajectoryEvaluationTrace.from_readable_spans(
                agent_run_history
            )
            agent_run_history = str(trajectory_trace.spans)
        else:
            agent_run_history = str(agent_run_history)

        formatted_prompt = formatted_prompt.replace(
            self.agent_run_history_placeholder,
            agent_run_history,
        )

        return formatted_prompt

    async def _get_llm_response(self, evaluation_prompt: str) -> LLMResponse:
        """Get response from the LLM using universal function calling.

        Args:
            evaluation_prompt: The formatted prompt to send to the LLM

        Returns:
            LLMResponse with score and justification
        """
        assert self.llm, "LLM should be initialized before calling this method."

        model = self.model
        if model.endswith(COMMUNITY_agents_SUFFIX):
            model = model.replace(COMMUNITY_agents_SUFFIX, "")

        # Create evaluation tool for function calling (works across all models)
        evaluation_tool = create_evaluation_tool()
        tool_choice = RequiredToolChoice()

        # Prepare the request with function calling
        request_data = {
            "model": model,
            "messages": [{"role": "user", "content": evaluation_prompt}],
            "tools": [evaluation_tool],
            "tool_choice": tool_choice,
        }

        response = await self.llm.chat_completions(**request_data)
        return extract_tool_call_response(response, model)



"""Tests for JobsService PagedResult pagination."""

import pytest

from uipath.platform.common.paging import PagedResult
from uipath.platform.orchestrator.job import Job


class TestJobsListPagination:
    """Test list() pagination with PagedResult."""

    def test_list_returns_paged_result(
        self, jobs_service, httpx_mock, base_url, org, tenant
    ):
        """Test that list() returns PagedResult[Job]."""
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/odata/Jobs?%24skip=0&%24top=100",
            json={"value": [{"Key": "job-1", "Id": 1, "State": "Successful"}]},
        )

        result = jobs_service.list()

        assert isinstance(result, PagedResult)
        assert len(result.items) == 1
        assert isinstance(result.items[0], Job)
        assert result.skip == 0
        assert result.top == 100

    def test_list_has_more_true_when_full_page(
        self, jobs_service, httpx_mock, base_url, org, tenant
    ):
        """Test has_more=True when page is full."""
        jobs = [{"Key": f"job-{i}", "Id": i, "State": "Successful"} for i in range(100)]
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/odata/Jobs?%24skip=0&%24top=100",
            json={"value": jobs},
        )

        result = jobs_service.list(top=100)

        assert result.has_more is True
        assert len(result.items) == 100

    def test_list_has_more_false_when_partial_page(
        self, jobs_service, httpx_mock, base_url, org, tenant
    ):
        """Test has_more=False when page is partial."""
        jobs = [{"Key": f"job-{i}", "Id": i, "State": "Successful"} for i in range(50)]
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/odata/Jobs?%24skip=0&%24top=100",
            json={"value": jobs},
        )

        result = jobs_service.list(top=100)

        assert result.has_more is False
        assert len(result.items) == 50

    def test_list_with_filter(self, jobs_service, httpx_mock, base_url, org, tenant):
        """Test list with OData filter."""
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/odata/Jobs?%24filter=State+eq+%27Successful%27&%24skip=0&%24top=100",
            json={"value": [{"Key": "job-1", "Id": 1, "State": "Successful"}]},
        )

        result = jobs_service.list(filter="State eq 'Successful'")

        assert len(result.items) == 1
        assert result.items[0].state == "Successful"

    def test_list_manual_pagination(
        self, jobs_service, httpx_mock, base_url, org, tenant
    ):
        """Test manual pagination across multiple pages."""
        # First page
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/odata/Jobs?%24skip=0&%24top=10",
            json={"value": [{"Key": f"job-{i}", "Id": i} for i in range(10)]},
        )
        # Second page
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/odata/Jobs?%24skip=10&%24top=10",
            json={"value": [{"Key": f"job-{i}", "Id": i} for i in range(10, 15)]},
        )

        # Fetch first page
        page1 = jobs_service.list(skip=0, top=10)
        assert len(page1.items) == 10
        assert page1.has_more is True

        # Fetch second page
        page2 = jobs_service.list(skip=10, top=10)
        assert len(page2.items) == 5
        assert page2.has_more is False


class TestJobsListValidation:
    """Test parameter validation for list()."""

    def test_list_skip_exceeds_maximum(self, jobs_service):
        """Test error when skip > MAX_SKIP_OFFSET."""
        with pytest.raises(
            ValueError, match=r"skip must be <= 10000.*requested: 10001"
        ):
            jobs_service.list(skip=10001)

    def test_list_top_exceeds_maximum(self, jobs_service):
        """Test error when top > MAX_PAGE_SIZE."""
        with pytest.raises(ValueError, match=r"top must be <= 1000.*requested: 1001"):
            jobs_service.list(top=1001)

    def test_list_uses_shared_validation(self, jobs_service):
        """Test that list() uses shared validation utility."""
        with pytest.raises(ValueError, match="skip must be >= 0"):
            jobs_service.list(skip=-1)

        with pytest.raises(ValueError, match="top must be >= 1"):
            jobs_service.list(top=0)

    def test_list_skip_at_boundary(
        self, jobs_service, httpx_mock, base_url, org, tenant
    ):
        """Test that skip=10000 is allowed."""
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/odata/Jobs?%24skip=10000&%24top=100",
            json={"value": []},
        )

        result = jobs_service.list(skip=10000)
        assert result is not None

    def test_list_top_at_boundary(
        self, jobs_service, httpx_mock, base_url, org, tenant
    ):
        """Test that top=1000 is allowed."""
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/odata/Jobs?%24skip=0&%24top=1000",
            json={"value": []},
        )

        result = jobs_service.list(top=1000)
        assert result is not None


class TestJobsListAsync:
    """Test async version of list()."""

    @pytest.mark.asyncio
    async def test_list_async_returns_paged_result(
        self, jobs_service, httpx_mock, base_url, org, tenant
    ):
        """Test that list_async() returns PagedResult[Job]."""
        httpx_mock.add_response(
            url=f"{base_url}{org}{tenant}/orchestrator_/odata/Jobs?%24skip=0&%24top=100",
            json={"value": [{"Key": "job-1", "Id": 1}]},
        )

        result = await jobs_service.list_async()

        assert isinstance(result, PagedResult)
        assert len(result.items) == 1



import logging
from enum import Enum
from typing import Any, Union

from opentelemetry.sdk.trace import ReadableSpan
from pydantic import BaseModel, ConfigDict, SkipValidation, model_validator

from uipath._cli._evals._models._evaluation_set import EvaluationItem
from uipath.eval.evaluators import BaseEvaluator
from uipath.eval.models import EvalItemResult


class EvaluationEvents(str, Enum):
    CREATE_EVAL_SET_RUN = "create_eval_set_run"
    CREATE_EVAL_RUN = "create_eval_run"
    UPDATE_EVAL_SET_RUN = "update_eval_set_run"
    UPDATE_EVAL_RUN = "update_eval_run"


class EvalSetRunCreatedEvent(BaseModel):
    execution_id: str
    entrypoint: str
    eval_set_id: str
    eval_set_run_id: str | None = None
    no_of_evals: int
    # skip validation to avoid abstract class instantiation
    evaluators: SkipValidation[list[BaseEvaluator[Any, Any, Any]]]


class EvalRunCreatedEvent(BaseModel):
    execution_id: str
    eval_item: EvaluationItem


class EvalItemExceptionDetails(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    runtime_exception: bool = False
    exception: Exception


class EvalRunUpdatedEvent(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    execution_id: str
    eval_item: EvaluationItem
    eval_results: list[EvalItemResult]
    success: bool
    agent_output: Any
    agent_execution_time: float
    spans: list[ReadableSpan]
    logs: list[logging.LogRecord]
    exception_details: EvalItemExceptionDetails | None = None

    @model_validator(mode="after")
    def validate_exception_details(self):
        if not self.success and self.exception_details is None:
            raise ValueError("exception_details must be provided when success is False")
        return self


class EvalSetRunUpdatedEvent(BaseModel):
    execution_id: str
    evaluator_scores: dict[str, float]
    success: bool = True


ProgressEvent = Union[
    EvalSetRunCreatedEvent,
    EvalRunCreatedEvent,
    EvalRunUpdatedEvent,
    EvalSetRunUpdatedEvent,
]



import logging
from dataclasses import dataclass


logger = logging.getLogger(__name__)


@dataclass
class EchoIn:
    message: str
    repeat: int | None = 1
    prefix: str | None = None


@dataclass
class EchoOut:
    message: str


def main(input: EchoIn) -> EchoOut:
    result = []

    for _ in range(input.repeat):
        line = input.message
        if input.prefix:
            line = f"{input.prefix}: {line}"
        result.append(line)

    return EchoOut(message="\n".join(result))



"""Base class for all output evaluator configurations."""

import json
from typing import Any, TypeVar, Union

from pydantic import Field

from ..models import AgentExecution
from ..models.models import UiPathEvaluationError, UiPathEvaluationErrorCategory
from .base_evaluator import (
    BaseEvaluationCriteria,
    BaseEvaluator,
    BaseEvaluatorConfig,
    BaseEvaluatorJustification,
)


class OutputEvaluationCriteria(BaseEvaluationCriteria):
    """Base class for all output evaluation criteria."""

    expected_output: dict[str, Any] | str = Field(..., alias="expectedOutput")


T = TypeVar("T", bound=BaseEvaluationCriteria)
T_OutputCriteria = TypeVar("T_OutputCriteria", bound=OutputEvaluationCriteria)


class OutputEvaluatorConfig(BaseEvaluatorConfig[T]):
    """Base class for all output evaluator configurations.

    Generic over T to allow subclasses to define their own
    specific output evaluation criteria types while maintaining type safety.
    """

    target_output_key: str = Field(
        default="*", description="Key to extract output from agent execution"
    )


C = TypeVar("C", bound=OutputEvaluatorConfig[Any])
J = TypeVar("J", bound=Union[str, None, BaseEvaluatorJustification])


class BaseOutputEvaluator(BaseEvaluator[T, C, J]):
    """Abstract base class for all output evaluators.

    Generic Parameters:
        T_OutputCriteria: The output evaluation criteria type
        C: The output evaluator config type (bound to OutputEvaluatorConfig[T_OutputCriteria])
        J: The justification type
    """

    def _get_actual_output(self, agent_execution: AgentExecution) -> Any:
        """Get the actual output from the agent execution."""
        if self.evaluator_config.target_output_key != "*" and isinstance(
            agent_execution.agent_output, dict
        ):
            try:
                return agent_execution.agent_output[
                    self.evaluator_config.target_output_key
                ]
            except KeyError as e:
                raise UiPathEvaluationError(
                    code="TARGET_OUTPUT_KEY_NOT_FOUND",
                    title="Target output key not found in actual output",
                    detail=f"Error: {e}",
                    category=UiPathEvaluationErrorCategory.USER,
                ) from e
        return agent_execution.agent_output

    def _get_full_expected_output(self, evaluation_criteria: T) -> Any:
        """Get the full expected output from the evaluation criteria."""
        raise UiPathEvaluationError(
            code="NOT_IMPLEMENTED",
            title="This method was not implemented by the subclass.",
            detail="This method was not implemented by the subclass.",
            category=UiPathEvaluationErrorCategory.SYSTEM,
        )

    def _get_expected_output(self, evaluation_criteria: T) -> Any:
        """Load the expected output from the evaluation criteria."""
        expected_output = self._get_full_expected_output(evaluation_criteria)
        if self.evaluator_config.target_output_key != "*":
            if isinstance(expected_output, str):
                try:
                    expected_output = json.loads(expected_output)
                except json.JSONDecodeError as e:
                    raise UiPathEvaluationError(
                        code="INVALID_EXPECTED_OUTPUT",
                        title="When target output key is not '*', expected output must be a dictionary or a valid JSON string",
                        detail=f"Error: {e}",
                        category=UiPathEvaluationErrorCategory.USER,
                    ) from e
            try:
                expected_output = expected_output[
                    self.evaluator_config.target_output_key
                ]
            except KeyError as e:
                raise UiPathEvaluationError(
                    code="TARGET_OUTPUT_KEY_NOT_FOUND",
                    title="Target output key not found in expected output",
                    detail=f"Error: {e}",
                    category=UiPathEvaluationErrorCategory.USER,
                ) from e
        return expected_output


class OutputEvaluator(BaseOutputEvaluator[T_OutputCriteria, C, J]):
    """Abstract base class for all output evaluators.

    Generic Parameters:
        T_OutputCriteria: The output evaluation criteria type
        C: The output evaluator config type (bound to OutputEvaluatorConfig[T_OutputCriteria])
        J: The justification type
    """

    def _get_full_expected_output(self, evaluation_criteria: T_OutputCriteria) -> Any:
        """Get the full expected output from the evaluation criteria."""
        return evaluation_criteria.expected_output



import json
from typing import Any

import click


def _get_option_info(param: click.Parameter) -> dict[str, Any]:
    info: dict[str, Any] = {"name": param.name}

    if isinstance(param, click.Option):
        info["flags"] = list(param.opts)
        if param.secondary_opts:
            info["flags"].extend(param.secondary_opts)
        if param.required:
            info["required"] = True
        if param.is_flag:
            info["is_flag"] = True
        if param.default is not None and not param.is_flag:
            if "Sentinel" in type(param.default).__name__:
                info["default"] = "unset"
            else:
                info["default"] = param.default
        if param.help:
            info["description"] = param.help

    if hasattr(param.type, "name"):
        info["type"] = param.type.name

    if isinstance(param.type, click.Choice):
        info["choices"] = list(param.type.choices)

    return info


def _get_command_info(
    cmd: click.Command, name: str, parent_ctx: click.Context | None = None
) -> dict[str, Any]:
    """Extract metadata from a Click command."""
    info: dict[str, Any] = {"name": name}

    if cmd.help:
        help_text = cmd.help.split("\n\n")[0].replace("\b", "").strip()
        info["description"] = help_text

    options = [
        _get_option_info(p)
        for p in cmd.params
        if isinstance(p, click.Option) and p.name != "help" and not p.hidden
    ]
    if options:
        info["options"] = options

    arguments = [
        {"name": p.name, "required": p.required}
        for p in cmd.params
        if isinstance(p, click.Argument)
    ]
    if arguments:
        info["arguments"] = arguments

    if isinstance(cmd, click.Group):
        sub_ctx = click.Context(cmd, parent=parent_ctx, info_name=name)
        subcommands = []
        for sub_name in cmd.list_commands(sub_ctx):
            sub_cmd = cmd.get_command(sub_ctx, sub_name)
            if sub_cmd:
                subcommands.append(_get_command_info(sub_cmd, sub_name, sub_ctx))
        if subcommands:
            info["subcommands"] = subcommands

    return info


def get_help_json(group: click.Group, ctx: click.Context, version: str) -> str:
    """Generate JSON help for the CLI.

    Args:
        group: The root CLI group
        ctx: Click context
        version: CLI version string

    Returns:
        JSON string with CLI structure
    """
    result: dict[str, Any] = {
        "name": ctx.info_name or "uipath",
        "version": version,
    }

    if group.help:
        result["description"] = group.help.split("\n\n")[0].replace("\b", "").strip()

    options = [
        _get_option_info(p)
        for p in group.params
        if isinstance(p, click.Option) and p.name != "help" and not p.hidden
    ]
    if options:
        result["options"] = options

    commands = []
    for cmd_name in group.list_commands(ctx):
        cmd = group.get_command(ctx, cmd_name)
        if cmd:
            commands.append(_get_command_info(cmd, cmd_name, ctx))

    result["commands"] = commands

    return json.dumps(result, indent=2, default=str)



"""Assets service commands for UiPath CLI.

Assets are key-value pairs stored in Orchestrator that can be used to store
configuration data, credentials, and other settings used by automation processes.
"""

import click

from .._utils._service_base import (
    ServiceCommandBase,
    common_service_options,
    service_command,
)


@click.group()
def assets():
    r"""Manage UiPath assets.

    Assets are key-value pairs that store configuration data, credentials,
    and settings used by automation processes.

    \b
    Examples:
        # List all assets in a folder
        uipath assets list --folder-path "Shared"

        # List with filter
        uipath assets list --filter "ValueType eq 'Text'"

        # List with ordering
        uipath assets list --orderby "Name asc"
    """
    pass


@assets.command(name="list")
@click.option("--filter", "filter_", help="OData $filter expression")
@click.option("--orderby", help="OData $orderby expression")
@click.option(
    "--top",
    type=click.IntRange(min=1, max=1000),
    default=100,
    help="Maximum number of items to return (default: 100, max: 1000)",
)
@click.option(
    "--skip",
    type=click.IntRange(min=0),
    default=0,
    help="Number of items to skip",
)
@common_service_options
@service_command
def list_assets(
    ctx,
    filter_,
    orderby,
    top,
    skip,
    folder_path,
    folder_key,
    format,
    output,
):
    r"""List assets in a folder.

    \b
    Examples:
        uipath assets list
        uipath assets list --folder-path "Shared"
        uipath assets list --filter "ValueType eq 'Text'"
        uipath assets list --filter "Name eq 'MyAsset'"
        uipath assets list --orderby "Name asc"
        uipath assets list --top 50 --skip 100
    """
    client = ServiceCommandBase.get_client(ctx)

    result = client.assets.list(
        folder_path=folder_path,
        folder_key=folder_key,
        filter=filter_,
        orderby=orderby,
        top=top,
        skip=skip,
    )

    return list(result.items)



import click

from ._auth._auth_service import AuthService
from ._utils._common import environment_options
from ._utils._console import ConsoleLogger

console = ConsoleLogger()


@click.command()
@environment_options
@click.option(
    "-f",
    "--force",
    is_flag=True,
    required=False,
    default=False,
    help="Force new token",
)
@click.option(
    "--client-id",
    required=False,
    help="Client ID for client credentials authentication (unattended mode)",
)
@click.option(
    "--client-secret",
    required=False,
    help="Client secret for client credentials authentication (unattended mode)",
)
@click.option(
    "--base-url",
    required=False,
    help="Base URL for the UiPath tenant instance (required for client credentials)",
)
@click.option(
    "--tenant",
    required=False,
    help="Tenant name within UiPath Automation Cloud",
)
@click.option(
    "--scope",
    required=False,
    default="OR.Execution",
    help="Space-separated list of OAuth scopes to request (e.g., 'OR.Execution OR.Queues'). Defaults to 'OR.Execution'",
)
def auth(
    environment: str,
    force: bool = False,
    client_id: str | None = None,
    client_secret: str | None = None,
    base_url: str | None = None,
    tenant: str | None = None,
    scope: str | None = None,
):
    """Authenticate with UiPath Cloud Platform.

    The domain for authentication is determined by the UIPATH_URL environment variable if set.
    Otherwise, it can be specified with --cloud (default), --staging, or --alpha flags.

    Interactive mode (default): Opens browser for OAuth authentication.

    Unattended mode: Use --client-id, --client-secret, --base-url and --scope for client credentials flow.

    Network options:

    - Set HTTP_PROXY/HTTPS_PROXY/NO_PROXY environment variables for proxy configuration
    - Set REQUESTS_CA_BUNDLE to specify a custom CA bundle for SSL verification
    - Set UIPATH_DISABLE_SSL_VERIFY to disable SSL verification (not recommended)
    """
    auth_service = AuthService(
        environment=environment,
        force=force,
        client_id=client_id,
        client_secret=client_secret,
        base_url=base_url,
        tenant=tenant,
        scope=scope,
    )
    with console.spinner("Authenticating with UiPath ..."):
        try:
            auth_service.authenticate()
            console.success(
                "Authentication successful.",
            )
        except KeyboardInterrupt:
            console.error(
                "Authentication cancelled by user.",
            )



"""UiPath Functions Runtime - factory and runtime for function-based execution."""

from uipath.runtime import UiPathRuntimeFactoryRegistry

from .factory import UiPathFunctionsRuntimeFactory
from .runtime import UiPathFunctionsRuntime


def register_default_runtime_factory():
    """Register the default functions factory."""
    UiPathRuntimeFactoryRegistry.register(
        "uipath",
        factory_callable=lambda context: UiPathFunctionsRuntimeFactory(
            config_path="uipath.json",
        ),
        config_file="uipath.json",
    )
    UiPathRuntimeFactoryRegistry.set_default("uipath")


__all__ = [
    "UiPathFunctionsRuntimeFactory",
    "UiPathFunctionsRuntime",
    "register_default_runtime_factory",
]



"""Unit tests for eval resume flow to ensure UiPathExecuteOptions is passed correctly."""

from pathlib import Path
from unittest.mock import AsyncMock, patch

import pytest
from uipath.core.tracing import UiPathTraceManager
from uipath.runtime import (
    UiPathExecuteOptions,
    UiPathRuntimeFactoryProtocol,
    UiPathRuntimeProtocol,
    UiPathRuntimeResult,
    UiPathRuntimeStatus,
)

from uipath._cli._evals._runtime import UiPathEvalContext, UiPathEvalRuntime
from uipath._events._event_bus import EventBus

# ============================================================================
# Direct unit tests using mocks to verify the specific code path we changed
# ============================================================================
#
# These tests directly verify that UiPathExecuteOptions is being passed correctly
# in the execute_runtime method, which is the specific code path we modified.
#


@pytest.mark.asyncio
async def test_execute_runtime_method_passes_options_with_resume_false():
    """Direct test of execute_runtime method to verify UiPathExecuteOptions(resume=False) is passed."""
    # Arrange
    from uipath._cli._evals._models._evaluation_set import EvaluationItem

    event_bus = EventBus()
    trace_manager = UiPathTraceManager()
    context = UiPathEvalContext()
    context.eval_set = str(
        Path(__file__).parent / "evals" / "eval-sets" / "default.json"
    )
    context.resume = False  # Test resume=False

    # Create a mock runtime that will be wrapped
    mock_runtime = AsyncMock(spec=UiPathRuntimeProtocol)
    mock_runtime.execute = AsyncMock(
        return_value=UiPathRuntimeResult(
            output={"result": "success"}, status=UiPathRuntimeStatus.SUCCESSFUL
        )
    )

    # Create a mock factory
    mock_factory = AsyncMock(spec=UiPathRuntimeFactoryProtocol)
    mock_factory.new_runtime = AsyncMock(return_value=mock_runtime)

    eval_runtime = UiPathEvalRuntime(context, mock_factory, trace_manager, event_bus)

    eval_item = EvaluationItem(
        id="test-eval",
        name="Test Evaluation",
        inputs={"foo": "bar"},
        evaluation_criterias={},
    )

    #  Act - Call execute_runtime directly
    with patch(
        "uipath._cli._evals._runtime.UiPathExecutionRuntime"
    ) as mock_execution_runtime_class:
        # Set up the mock to capture the execute call
        mock_execution_runtime_instance = AsyncMock()
        mock_execution_runtime_instance.execute = AsyncMock(
            return_value=UiPathRuntimeResult(
                output={"result": "success"}, status=UiPathRuntimeStatus.SUCCESSFUL
            )
        )
        mock_execution_runtime_class.return_value = mock_execution_runtime_instance

        await eval_runtime.execute_runtime(
            eval_item=eval_item, execution_id="test-exec-id", runtime=mock_runtime
        )

        # Assert - Verify that execute was called with UiPathExecuteOptions(resume=False)
        assert mock_execution_runtime_instance.execute.called
        call_args = mock_execution_runtime_instance.execute.call_args

        # Extract the options argument
        options = call_args.kwargs.get("options") or (
            call_args.args[1] if len(call_args.args) > 1 else None
        )

        # Assert that options were passed and resume=False
        assert options is not None, "UiPathExecuteOptions should be passed explicitly"
        assert isinstance(options, UiPathExecuteOptions)
        assert options.resume is False, (
            "resume should be False when context.resume=False"
        )


@pytest.mark.asyncio
async def test_execute_runtime_method_passes_options_with_resume_true():
    """Direct test of execute_runtime method to verify UiPathExecuteOptions(resume=True) is passed."""
    # Arrange
    from uipath._cli._evals._models._evaluation_set import EvaluationItem

    event_bus = EventBus()
    trace_manager = UiPathTraceManager()
    context = UiPathEvalContext()
    context.eval_set = str(
        Path(__file__).parent / "evals" / "eval-sets" / "default.json"
    )
    context.resume = True  # Test resume=True

    # Create a mock runtime
    mock_runtime = AsyncMock(spec=UiPathRuntimeProtocol)
    mock_runtime.execute = AsyncMock(
        return_value=UiPathRuntimeResult(
            output={"result": "success"}, status=UiPathRuntimeStatus.SUCCESSFUL
        )
    )

    # Create a mock factory
    mock_factory = AsyncMock(spec=UiPathRuntimeFactoryProtocol)
    mock_factory.new_runtime = AsyncMock(return_value=mock_runtime)

    eval_runtime = UiPathEvalRuntime(context, mock_factory, trace_manager, event_bus)

    eval_item = EvaluationItem(
        id="test-eval",
        name="Test Evaluation",
        inputs={"foo": "bar"},
        evaluation_criterias={},
    )

    # Act - Call execute_runtime directly
    with patch(
        "uipath._cli._evals._runtime.UiPathExecutionRuntime"
    ) as mock_execution_runtime_class:
        # Set up the mock to capture the execute call
        mock_execution_runtime_instance = AsyncMock()
        mock_execution_runtime_instance.execute = AsyncMock(
            return_value=UiPathRuntimeResult(
                output={"result": "success"}, status=UiPathRuntimeStatus.SUCCESSFUL
            )
        )
        mock_execution_runtime_class.return_value = mock_execution_runtime_instance

        await eval_runtime.execute_runtime(
            eval_item=eval_item, execution_id="test-exec-id", runtime=mock_runtime
        )

        # Assert - Verify that execute was called with UiPathExecuteOptions(resume=True)
        assert mock_execution_runtime_instance.execute.called
        call_args = mock_execution_runtime_instance.execute.call_args

        # Extract the options argument
        options = call_args.kwargs.get("options") or (
            call_args.args[1] if len(call_args.args) > 1 else None
        )

        # Assert that options were passed and resume=True
        assert options is not None, "UiPathExecuteOptions should be passed explicitly"
        assert isinstance(options, UiPathExecuteOptions)
        assert options.resume is True, "resume should be True when context.resume=True"


@pytest.mark.asyncio
async def test_resume_with_multiple_evaluations_raises_error():
    """Test that resume mode with multiple evaluations raises a ValueError."""
    # Arrange
    event_bus = EventBus()
    trace_manager = UiPathTraceManager()
    context = UiPathEvalContext()
    context.eval_set = str(
        Path(__file__).parent / "evals" / "eval-sets" / "multiple-evals.json"
    )
    context.resume = True  # Enable resume mode

    # Create a mock factory
    mock_factory = AsyncMock(spec=UiPathRuntimeFactoryProtocol)
    mock_runtime = AsyncMock(spec=UiPathRuntimeProtocol)
    mock_factory.new_runtime = AsyncMock(return_value=mock_runtime)

    eval_runtime = UiPathEvalRuntime(
        context=context,
        factory=mock_factory,
        event_bus=event_bus,
        trace_manager=trace_manager,
    )

    # Act & Assert
    with pytest.raises(
        ValueError,
        match=r"Resume mode is not supported with multiple evaluations.*Found 2 evaluations",
    ):
        await eval_runtime.initiate_evaluation(mock_runtime)



from typing import TypedDict


class AuthConfig(TypedDict):
    """TypedDict for auth_config.json structure."""

    client_id: str
    port: int
    redirect_uri: str
    scope: str


class AccessTokenData(TypedDict):
    """TypedDict for access token data structure."""

    sub: str
    prt_id: str
    client_id: str
    exp: float


class TenantInfo(TypedDict):
    """TypedDict for tenant info structure."""

    name: str
    id: str


class OrganizationInfo(TypedDict):
    """TypedDict for organization info structure."""

    id: str
    name: str


class TenantsAndOrganizationInfoResponse(TypedDict):
    """TypedDict for tenants and organization info response structure."""

    tenants: list[TenantInfo]
    organization: OrganizationInfo



"""Models for Resource Catalog service."""

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, ConfigDict, Field


class ResourceType(str, Enum):
    """Resource type."""

    ASSET = "asset"
    BUCKET = "bucket"
    MACHINE = "machine"
    TRIGGER = "trigger"
    PROCESS = "process"
    PACKAGE = "package"
    LIBRARY = "library"
    INDEX = "index"
    APP = "app"
    CONNECTION = "connection"
    CONNECTOR = "connector"
    MCP_SERVER = "mcpserver"
    QUEUE = "queue"

    @classmethod
    def from_string(cls, value: str) -> "ResourceType":
        """Create a ResourceType instance from a string value.

        Args:
            value: String value to convert to ResourceType

        Returns:
            ResourceType: The matching ResourceType enum member

        Raises:
            ValueError: If the value doesn't match any ResourceType
        """
        lower_value = value.lower()
        for member in cls:
            if member.value == lower_value:
                return member

        available = ", ".join([f"'{member.value}'" for member in cls])
        raise ValueError(
            f"'{value}' is not a valid ResourceType. Available options: {available}"
        )


class Tag(BaseModel):
    """Tag model for resources."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )

    key: str
    display_name: str = Field(alias="displayName")
    name: str
    display_value: str = Field(alias="displayValue")
    value: str
    type: str
    account_key: str = Field(alias="accountKey")
    tenant_key: Optional[str] = Field(None, alias="tenantKey")
    user_key: Optional[str] = Field(None, alias="userKey")


class Folder(BaseModel):
    """Folder model for resources."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )

    id: int
    key: str
    display_name: str = Field(alias="displayName")
    code: str
    fully_qualified_name: str = Field(alias="fullyQualifiedName")
    timestamp: str
    tenant_key: str = Field(alias="tenantKey")
    account_key: str = Field(alias="accountKey")
    user_key: Optional[str] = Field(None, alias="userKey")
    type: str
    path: str
    permissions: Optional[List[str]] = Field(default_factory=list)


class Resource(BaseModel):
    """Resource model from Resource Catalog."""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
    )

    resource_key: str = Field(alias="entityKey")
    name: str
    description: Optional[str] = None
    resource_type: str = Field(alias="entityType")
    tags: Optional[List[Tag]] = Field(default_factory=list)
    folders: List[Folder] = Field(default_factory=list)
    linked_folders_count: int = Field(0, alias="linkedFoldersCount")
    source: Optional[str] = None
    scope: str
    search_state: str = Field(alias="searchState")
    timestamp: str
    folder_key: Optional[str] = Field(None, alias="folderKey")
    folder_keys: List[str] = Field(default_factory=list, alias="folderKeys")
    tenant_key: Optional[str] = Field(None, alias="tenantKey")
    account_key: str = Field(alias="accountKey")
    user_key: Optional[str] = Field(None, alias="userKey")
    dependencies: Optional[list[str]] = Field(default_factory=list)
    custom_data: Optional[str] = Field(None, alias="customData")
    resource_sub_type: Optional[str] = Field(None, alias="entitySubType")


class ResourceSearchResponse(BaseModel):
    """Response model for resource search API."""

    count: int
    value: List[Resource]
